"""Contains a container for the cell output area."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from pathlib import PurePath
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.layout.containers import DynamicContainer, HSplit, to_container
from prompt_toolkit.widgets.base import Box

from euporie.core.app import get_app
from euporie.core.convert.base import BASE64_FORMATS, MIME_FORMATS, find_route
from euporie.core.widgets.display import Display
from euporie.core.widgets.tree import JsonView

if TYPE_CHECKING:
    from typing import Any, Optional, Protocol, Type, TypeVar

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.comm.base import KernelTab

    KTParent = TypeVar("KTParent", bound=KernelTab)

    class OutputParent(Protocol[KTParent]):
        """An output's parent."""

        kernel_tab: KTParent

        def refresh(self, now: "bool" = True) -> "None":
            """Update the parent container."""
            ...


log = logging.getLogger(__name__)


class CellOutputElement(metaclass=ABCMeta):
    """Base class for the various types of cell outputs (display data or widgets)."""

    data: "Any"

    def __init__(
        self,
        mime: "str",
        data: "Any",
        metadata: "dict",
        parent: "Optional[OutputParent]",
    ) -> "None":
        """Create a new instances of the output element.

        Args:
            mime: The mime-type of the data to display
            data: The data to display
            metadata: Any metadata relating to the data
            parent: The cell the output-element is attached to

        """
        ...

    def scroll_left(self) -> "None":
        """Scrolls the output left."""
        pass

    def scroll_right(self) -> "None":
        """Scrolls the output right."""
        pass

    @abstractmethod
    def __pt_container__(self) -> "AnyContainer":
        """Return the container representing the cell output element."""
        ...


class CellOutputDataElement(CellOutputElement):
    """A cell output element which display data."""

    def __init__(
        self,
        mime: "str",
        data: "Any",
        metadata: "dict",
        parent: "Optional[OutputParent]",
    ) -> "None":
        """Create a new data output element instance.

        Args:
            mime: The mime-type of the data to display
            data: The data to display
            metadata: Any metadata relating to the data
            parent: The cell the output-element is attached to
        """
        self.parent = parent
        self._data = data

        # Get foreground and background colors
        fg_color = None
        if parent is not None:
            fg_color = get_app().color_palette.fg.base_hex
        bg_color = {"light": "#FFFFFF", "dark": "#000000"}.get(
            str(metadata.get("needs_background"))
        )

        # Get internal format
        format_ = "ansi"
        mime_path = PurePath(mime)
        for mime_type, data_format in MIME_FORMATS.items():
            if mime_path.match(mime_type):
                if data_format in BASE64_FORMATS:
                    data_format = f"base64-{data_format}"
                if find_route(data_format, "formatted_text") is not None:
                    format_ = data_format
                    break

        self.container = Display(
            data=data,
            format_=format_,
            fg_color=fg_color,
            bg_color=bg_color,
            px=metadata.get("width"),
            py=metadata.get("height"),
            focusable=False,
            focus_on_click=False,
            wrap_lines=False,
            always_hide_cursor=True,
            style=f"class:cell.output.element.data class:mime.{mime.replace('/','.')}",
            scrollbar=False,
        )

    @property
    def data(self) -> "Any":
        """Return the control's display data."""
        return self._data

    @data.setter
    def data(self, value: "Any") -> "None":
        self._data = value
        self.container.data = value

    def scroll_left(self) -> "None":
        """Scrolls the output left."""
        self.container.window._scroll_left()

    def scroll_right(self) -> "None":
        """Scrolls the output right."""
        self.container.window._scroll_right()

    def __pt_container__(self) -> "AnyContainer":
        """Return the display container."""
        return self.container


class CellOutputWidgetElement(CellOutputElement):
    """A cell output element which displays ipywidgets."""

    def __init__(
        self,
        mime: "str",
        data: "dict[str, Any]",
        metadata: "dict",
        parent: "Optional[OutputParent]",
    ) -> "None":
        """Create a new widget output element instance.

        Args:
            mime: The mime-type of the data to display:
                ``application/vnd.jupyter.widget-view+json``
            data: The data to display (the widget model ID)
            metadata: Any metadata relating to the data
            parent: The parent container the output-element is attached to

        Raises:
            NotImplementedError: Raised when an ipywidget cannot be rendered
        """
        self.parent = parent
        self.comm_id = str(data.get("model_id"))

        if parent is not None and (comm := parent.kernel_tab.comms.get(self.comm_id)):
            self.container = Box(
                comm.new_view(parent),
                padding_left=0,
                padding_top=0,
                padding_bottom=0,
                style="class:ipywidget",
            )
        else:
            raise NotImplementedError

    def __pt_container__(self) -> "AnyContainer":
        """Return a box container which holds a view of an ipywidget."""
        return self.container


class CellOutputJsonElement(CellOutputElement):
    """A cell output element which displays JSON."""

    def __init__(
        self,
        mime: "str",
        data: "dict[str, Any]",
        metadata: "dict",
        parent: "Optional[OutputParent]",
    ) -> "None":
        """Create a new widget output element instance.

        Args:
            mime: The mime-type of the data to display: ``application/json``
            data: The data to display (a JSON object)
            metadata: Any metadata relating to the data
            parent: The parent container the output-element is attached to
        """
        self.parent = parent
        self.container = JsonView(
            data, title=metadata.get("root"), expanded=metadata.get("expanded", True)
        )

    def __pt_container__(self) -> "AnyContainer":
        """Return a box container which holds a view of an ipywidget."""
        return self.container


MIME_RENDERERS: "dict[str, Type[CellOutputElement]]" = {
    "application/vnd.jupyter.widget-view+json": CellOutputWidgetElement,
    "application/json": CellOutputJsonElement,
    "*": CellOutputDataElement,
}

MIME_ORDER = [
    "application/vnd.jupyter.widget-view+json",
    "application/json",
    "image/*",
    "text/html",
    "text/markdown",
    "text/x-markdown",
    "application/pdf",
    "text/latex",
    "text/x-python-traceback",
    "text/stderr",
    "text/*",
    "*",
]


def _calculate_mime_rank(mime_data: "tuple[str, Any]") -> "int":
    """Scores the richness of mime output types."""
    mime, data = mime_data
    for i, ranked_mime in enumerate(MIME_ORDER):
        # Uprank plain text with escape sequences
        if mime == "text/plain" and "\x1b[" in data:
            i -= 7
        if PurePath(mime).match(ranked_mime):
            return i
    else:
        return 999


class CellOutput:
    """Represents a single cell output.

    Capable of displaying multiple mime representations of the same data.

    TODO - allow the visible mime-type to be rotated.
    """

    def __init__(
        self, json: "dict[str, Any]", parent: "Optional[OutputParent]"
    ) -> "None":
        """Creates a new cell output instance.

        Args:
            json: The cell output's json
            parent: The output's parent container
        """
        # Select the first mime-type to render
        self.parent = parent
        self.json = json
        self.selected_mime = next(x for x in self.data)
        self._elements: "dict[str, CellOutputElement]" = {}

    @property
    def data(self) -> "dict[str, Any]":
        """Return dictionary of mime types and data for this output.

        This generates similarly structured data objects for markdown cells and text
        output streams.

        Returns:
            JSON dictionary mapping mimes type to representation data.

        """
        data = {}
        output_type = self.json.get("output_type", "unknown")
        if output_type == "stream":
            data = {f'stream/{self.json.get("name")}': self.json.get("text", "")}
        elif output_type == "error":
            ename = self.json.get("ename", "")
            evalue = self.json.get("evalue", "")
            traceback = "\n".join(self.json.get("traceback", ""))
            data = {"text/x-python-traceback": f"{ename}: {evalue}\n{traceback}"}
        else:
            data = self.json.get("data", {"text/plain": ""})
        return dict(sorted(data.items(), key=_calculate_mime_rank))

    def update(self) -> "None":
        """Update's the output by updating all child containers."""
        # log.debug("Updating %s", self)
        data = self.data
        for mime_type, element in self._elements.items():
            if mime_type in data:
                element.data = data[mime_type]
            else:
                del self._elements[mime_type]

    def make_element(self, mime: "str") -> "CellOutputElement":
        """Creates a container for the cell output mime-type if it doesn't exist.

        Args:
            mime: The mime-type for which to create an output element

        Returns:
            A :class:`OutputElement` container for the currently selected mime-type.
        """
        data = self.data
        for mime_pattern, OutputElement in MIME_RENDERERS.items():
            if PurePath(mime).match(mime_pattern):
                try:
                    element = OutputElement(
                        mime=mime,
                        data=data[mime],
                        metadata=self.json.get("metadata", {}).get(mime, {}),
                        parent=self.parent,
                    )
                except NotImplementedError:
                    self.selected_mime = mime = list(data.keys())[-1]
                    continue
                else:
                    return element
        return CellOutputDataElement(
            "text/plain", "(Cannot display output)", {}, self.parent
        )

    def get_element(self, mime: "str") -> "CellOutputElement":
        """Return the currently displayed cell element."""
        if mime not in self._elements:
            element = self.make_element(mime)
            self._elements[mime] = element
        else:
            element = self._elements[mime]
        return element

    @property
    def element(self) -> "CellOutputElement":
        """Get the element for the currently selected mime type."""
        return self.get_element(self.selected_mime)

    def __pt_container__(self):
        """Return the cell output container (an :class:`OutputElement`)."""
        return DynamicContainer(lambda: self.element)

    def scroll_left(self) -> "None":
        """Scrolls the currently visible output left."""
        self.element.scroll_left()

    def scroll_right(self) -> "None":
        """Scrolls the currently visible output right."""
        self.element.scroll_right()


class CellOutputArea:
    """An area below a cell where one or more cell outputs can be shown."""

    output_cache: "SimpleCache[str, CellOutput]" = SimpleCache()

    def __init__(
        self,
        json: "list[dict[str, Any]]",
        parent: "Optional[OutputParent]",
        style: "str" = "",
    ) -> "None":
        """Creates a new cell output area instance.

        Args:
            json: The cell's output json
            parent: The parent container the output belongs to
            style: Additional style to apply to the output

        """
        self._json: "list[dict[str, Any]]" = []
        self.parent = parent
        self.style = style

        self.display_json: "list[dict[str, Any]]" = []

        self.rendered_outputs: "list[CellOutput]" = []
        self.container = HSplit([], style=lambda: self.style)

        self.json = json

    @property
    def json(self) -> "Any":
        """Return the control's display data."""
        return self._json

    @json.setter
    def json(self, value: "Any") -> "None":
        # Reset if we have lost existing outputs
        if any(old_output_json not in value for old_output_json in self._json):
            self.reset()
        # Add any new outputs we do not already have
        for new_output_json in value:
            if new_output_json not in self._json:
                self.add_output(new_output_json, refresh=False)
        get_app().invalidate()

    def add_output(
        self, output_json: "dict[str, Any]", refresh: "bool" = True
    ) -> "None":
        """Add a new output to the output area."""
        # Update json
        self._json.append(output_json)
        # Update display json
        add_output = True
        if name := output_json.get("name"):
            for existing_output, rendered_output in zip(
                self.display_json, self.rendered_outputs
            ):
                if name == existing_output.get("name"):
                    existing_output["text"] += output_json.get("text", "")
                    rendered_output.update()
                    add_output = False
                    break
        if add_output:
            # Add a copy to the display json so the original does not get modified
            output_json_copy = dict(output_json)
            self.display_json.append(output_json_copy)
            # Create an output container with the copy
            output = CellOutput(output_json_copy, self.parent)
            self.rendered_outputs.append(output)
            self.container.children.append(to_container(output))
        if refresh:
            get_app().invalidate()

    def update(self) -> "None":
        """Update all existing outputs."""
        for output in self.rendered_outputs:
            output.update()

    def reset(self) -> "None":
        """Clears all outputs from the output area."""
        self.style = ""
        self._json.clear()
        self.display_json.clear()
        self.rendered_outputs.clear()
        self.container.children.clear()

    def scroll_left(self) -> "None":
        """Scrolls the outputs left."""
        for cell_output in self.rendered_outputs:
            cell_output.scroll_left()

    def scroll_right(self) -> "None":
        """Scrolls the outputs right."""
        for cell_output in self.rendered_outputs:
            cell_output.scroll_right()

    def __pt_container__(self) -> "AnyContainer":
        """Return the cell output area container (an :class:`HSplit`)."""
        return self.container

    def to_plain_text(self) -> "str":
        """Convert the contents of the output to plain text."""
        from prompt_toolkit.formatted_text.utils import to_plain_text

        outputs = []
        for cell_output in self.rendered_outputs:
            if isinstance(cell_output.element, CellOutputDataElement):
                for line in cell_output.element.container.control.get_rendered_lines(
                    width=88, height=99999999
                ):
                    outputs.append(to_plain_text(line))
        return "\n".join(outputs)



"""Contains function which convert data to rich format."""

from __future__ import annotations

from typing import TYPE_CHECKING

from euporie.core.convert.base import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from typing import Optional

    from rich.markdown import Markdown
    from upath import UPath


@register(
    from_="markdown",
    to="rich",
    filter_=have_modules("rich"),
)
def markdown_to_rich_py(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "Markdown":
    """Converts base64 encoded data to bytes."""
    from rich.markdown import Markdown

    return Markdown(
        data,
        # code_theme=str(config.syntax_theme),
        # inline_code_theme=str(config.syntax_theme),
    )



"""Defines custom controls which re-render on resize."""

from __future__ import annotations

import logging
import weakref
from abc import ABCMeta
from functools import partial
from math import ceil
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters.app import has_completions
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_width, split_lines
from prompt_toolkit.layout.containers import Float, Window
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl
from prompt_toolkit.layout.margins import ConditionalMargin
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, WritePosition
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import Event

from euporie.core.app import get_app
from euporie.core.commands import add_cmd
from euporie.core.convert.base import convert, find_route
from euporie.core.convert.utils import data_pixel_size, pixels_to_cell_size
from euporie.core.filters import display_has_focus, has_dialog, has_menus
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.margins import ScrollbarMargin
from euporie.core.terminal import tmuxify

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Optional, Type, Union

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding import KeyBindingsBase
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.layout.screen import Screen
    from upath import UPath


log = logging.getLogger(__name__)


class DisplayControl(UIControl):
    """A data formatter, which displays cell output data.

    It will attempt to display the data in the best way possible, and reacts to resize
    events - i.e. images are downscaled to fit, markdown is re-flowed, etc.
    """

    def __init__(
        self,
        data: "Any",
        format_: "str",
        path: "Optional[UPath]" = None,
        fg_color: "Optional[str]" = None,
        bg_color: "Optional[str]" = None,
        sizing_func: "Optional[Callable[[], tuple[int, float]]]" = None,
        focusable: "FilterOrBool" = False,
        focus_on_click: "FilterOrBool" = False,
    ) -> "None":
        """Creates a new data formatter control.

        Args:
            data: Raw cell output data
            format_: The conversion format of the data to render
            path: The path to the data's original location
            fg_color: The foreground colour to use when renderin this output
            bg_color: The background colour to use when renderin this output
            sizing_func: Function which returns the maximum width and aspect ratio of
                the output
            focusable: Whether the control can be focused
            focus_on_click: Whether to focus the control when clicked

        """
        self._data = data
        self.format_ = format_
        self.path = path
        self.fg_color = fg_color
        self.bg_color = bg_color
        self.focusable = to_filter(focusable)
        self.focus_on_click = to_filter(focus_on_click)
        self.key_bindings = load_registered_bindings(
            "euporie.core.widgets.display.Display"
        )
        self.app = get_app()

        self.on_cursor_position_changed = Event(self)
        self._cursor_position = Point(x=0, y=0)
        self.dy = 0

        self.sizing_func = sizing_func or (lambda: (0, 0))
        self._max_cols = 0
        self._aspect = 0.0

        self.rendered_lines: "list[StyleAndTextTuples]" = []
        self._format_cache: SimpleCache = SimpleCache(maxsize=50)
        self._content_cache: SimpleCache = SimpleCache(maxsize=50)
        self._size_cache: SimpleCache = SimpleCache(maxsize=1)

    def reset(self) -> "None":
        """Clear the display control's caches (required if the control's data changes)."""
        self._format_cache.clear()
        self._size_cache.clear()
        self._content_cache.clear()

    @property
    def data(self) -> "Any":
        """Return the control's display data."""
        return self._data

    @data.setter
    def data(self, value: "Any") -> "None":
        self._data = value
        self.reset()

    def get_key_bindings(self) -> "Optional[KeyBindingsBase]":
        """Return the control's key bindings."""
        return self.key_bindings

    @property
    def cursor_position(self) -> "Point":
        """Get the cursor position."""
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: "Point") -> "None":
        """Set the cursor position."""
        changed = self._cursor_position != value
        self._cursor_position = value
        if changed:
            self.on_cursor_position_changed.fire()

    def move_cursor_down(self) -> "None":
        """Moves the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=y + 1)

    def move_cursor_up(self) -> "None":
        """Moves the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=max(0, y - 1))

    def move_cursor_left(self) -> "None":
        """Moves the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=max(0, x - 1), y=y)

    def move_cursor_right(self) -> "None":
        """Moves the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x + 1, y=y)

    def is_focusable(self) -> "bool":
        """Determines if the current control is focusable."""
        return self.focusable()

    def size(self) -> "None":
        """Load the maximum cell width and apect ratio of the output."""
        self._max_cols, self._aspect = self._size_cache.get(
            self.app.render_counter, self.sizing_func
        )

    def hide(self) -> "None":
        """Hides the output from show."""
        pass

    @property
    def max_cols(self) -> "int":
        """Load the maximum width of the output in terminal columns."""
        self.size()
        return self._max_cols

    @property
    def aspect(self) -> "float":
        """Lazily load the aspect ratio of the output."""
        self.size()
        return self._aspect

    def preferred_width(self, max_available_width: "int") -> "Optional[int]":
        """Returns the width of the rendered content."""
        self.max_available_width = max_available_width
        return (
            min(self.max_cols, max_available_width)
            if self.max_cols
            else max_available_width
        )

    def get_rendered_lines(
        self, width: "int", height: "int"
    ) -> "list[StyleAndTextTuples]":
        """Render the output data."""
        return []

    def preferred_height(
        self,
        width: "int",
        max_available_height: "int",
        wrap_lines: "bool",
        get_line_prefix: Optional[GetLinePrefixCallable],
    ) -> "int":
        """Returns the number of lines in the rendered content."""
        if self.aspect:
            return ceil(min(width, self.max_cols) * self.aspect)
        else:
            self.rendered_lines = self.get_rendered_lines(width, max_available_height)
            return len(self.rendered_lines)

    def create_content(self, width: "int", height: "int") -> "UIContent":
        """Generates rendered output at a given size.

        Args:
            width: The desired output width
            height: The desired output height

        Returns:
            `UIContent` for the given output size.

        """
        cols = min(self.max_cols, width) if self.max_cols else width
        rows = ceil(cols * self.aspect) if self.aspect else height

        def get_content() -> "dict[str, Any]":
            rendered_lines = self.get_rendered_lines(width=cols, height=rows)
            self.rendered_lines = rendered_lines[:]
            line_count = len(rendered_lines)

            def get_line(i: "int") -> "StyleAndTextTuples":
                # Return blank lines if the renderer expects more content than we have
                line = []
                if i < line_count:
                    line += rendered_lines[i]
                # Add a space at the end, because that is a possible cursor position.
                # This is what PTK does, and fixes a nasty bug which took me ages to
                # track down the source of, where scrolling would stop working when the
                # cursor was on an empty line.
                line += [("", " ")]
                return line

            return {
                "get_line": get_line,
                "line_count": line_count,
                "menu_position": Point(0, 0),
            }

        # Re-render if the image width changes, or the terminal character size changes
        key = (cols, self.app.term_info.cell_size_px)
        return UIContent(
            cursor_position=self.cursor_position,
            **self._content_cache.get(key, get_content),
        )

    def mouse_handler(self, mouse_event: "MouseEvent") -> "NotImplementedOrNone":
        """Mouse handler for this control."""
        if self.focus_on_click() and mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.app.layout.current_control = self
            return None
        return NotImplemented

    def get_invalidate_events(self) -> "Iterable[Event[object]]":
        """Return the Window invalidate events."""
        # Whenever the cursor position changes, the UI has to be updated.
        yield self.on_cursor_position_changed

    @property
    def content_width(self) -> "int":
        """Return the maximum line length of the content."""
        return max(fragment_list_width(line) for line in self.rendered_lines)

    def close(self) -> "None":
        """Remove the displayed object entirely."""
        if not self.app.leave_graphics():
            self.hide()


class DisplayWindow(Window):
    """A window sub-class which can scroll left and right."""

    content: "DisplayControl"

    def _write_to_screen_at_index(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
    ) -> None:
        """Ensure the :attr:`horizontal_scroll` is recorded."""
        super()._write_to_screen_at_index(
            screen,
            mouse_handlers,
            write_position,
            parent_style,
            erase_bg,
        )
        # Set the horizontal scroll offset on the render info
        # TODO - fix this upstream
        if self.render_info:
            setattr(self.render_info, "horizontal_scroll", self.horizontal_scroll)

    def _scroll_right(self) -> None:
        """Scroll window right."""
        info = self.render_info
        if info is None:
            return
        content_width = self.content.content_width
        if self.horizontal_scroll < content_width - info.window_width:
            if info.cursor_position.y <= info.configured_scroll_offsets.right:
                self.content.move_cursor_right()

            self.horizontal_scroll += 1

    def _scroll_left(self) -> None:
        """Scroll window left."""
        info = self.render_info
        if info is None:
            return
        if self.horizontal_scroll > 0:
            if (
                info.cursor_position.x
                >= info.window_width - 1 - info.configured_scroll_offsets.left
            ):
                self.content.move_cursor_left()

            self.horizontal_scroll -= 1


class FormattedTextDisplayControl(DisplayControl):
    """A data formatter, which displays cell output data.

    It will attempt to display the data in the best way possible, and reacts to resize
    events - i.e. images are downscaled to fit, markdown is re-flowed, etc.
    """

    def get_rendered_lines(
        self, width: "int", height: "int"
    ) -> "list[StyleAndTextTuples]":
        """Get rendered lines from the cache, or generate them."""

        def render_lines() -> "list[StyleAndTextTuples]":
            """Renders the lines to display in the control."""
            return list(
                split_lines(
                    convert(
                        data=self.data,
                        from_=self.format_,
                        to="formatted_text",
                        cols=width,
                        rows=height,
                        fg=self.fg_color,
                        bg=self.bg_color,
                        path=self.path,
                    )
                )
            )

        # Re-render if the image width changes, or the terminal character size changes
        key = (width, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)


class GraphicControl(DisplayControl, metaclass=ABCMeta):
    """A base-class for display controls which render terminal graphics."""


class SixelGraphicControl(GraphicControl):
    """A graphic control which displays images as sixels."""

    def get_rendered_lines(
        self, width: "int", height: "int"
    ) -> "list[StyleAndTextTuples]":
        """Get rendered lines from the cache, or generate them."""

        def render_lines() -> "list[StyleAndTextTuples]":
            """Renders the lines to display in the control."""
            cmd = convert(
                data=self.data,
                from_=self.format_,
                to="sixel",
                cols=width,
                rows=height,
                fg=self.fg_color,
                bg=self.bg_color,
                path=self.path,
            )
            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and across by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            (
                                "[ZeroWidthEscape]",
                                f"\x1b[s\x1b[{height-1}A\x1b[{width}D",
                            ),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", tmuxify(cmd)),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)


class ItermGraphicControl(GraphicControl):
    """A graphic control which displays images using iTerm's graphics protocol."""

    def convert_data(self, rows: "int", cols: "int") -> "str":
        """Converts the graphic's data to base64 data."""
        if self.format_.startswith("base64-"):
            b64data = self.data
        else:
            b64data = convert(
                data=self.data,
                from_=self.format_,
                to="base64-png",
                cols=cols,
                rows=rows,
                fg=self.fg_color,
                bg=self.bg_color,
                path=self.path,
            )
        b64data = b64data.replace("\n", "").strip()
        return b64data

    def get_rendered_lines(
        self, width: "int", height: "int"
    ) -> "list[StyleAndTextTuples]":
        """Get rendered lines from the cache, or generate them."""

        def render_lines() -> "list[StyleAndTextTuples]":
            """Renders the lines to display in the control."""
            b64data = self.convert_data(cols=width, rows=height)
            cmd = f"\x1b]1337;File=inline=1;width={width}:{b64data}\a"
            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and across by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            (
                                "[ZeroWidthEscape]",
                                f"\x1b[s\x1b[{height-1}A\x1b[{width}D",
                            ),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", tmuxify(cmd)),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)


_kitty_image_count = 1


class KittyGraphicControl(GraphicControl):
    """A graphic control which displays images using Kitty's graphics protocol."""

    def __init__(
        self,
        data: "Any",
        format_: "str",
        path: "Optional[UPath]" = None,
        fg_color: "Optional[str]" = None,
        bg_color: "Optional[str]" = None,
        sizing_func: "Optional[Callable[[], tuple[int, float]]]" = None,
        focusable: "FilterOrBool" = False,
        focus_on_click: "FilterOrBool" = False,
    ) -> "None":
        """Create a new kitty graphic instance."""
        super().__init__(
            data=data,
            format_=format_,
            path=path,
            fg_color=fg_color,
            bg_color=bg_color,
            sizing_func=sizing_func,
            focusable=focusable,
            focus_on_click=focus_on_click,
        )
        self.kitty_image_id = 0
        self.loaded = False

    def convert_data(self, rows: "int", cols: "int") -> "str":
        """Converts the graphic's data to base64 data for kitty graphics protocol."""
        return convert(
            self.data,
            from_=self.format_,
            to="base64-png",
            cols=cols,
            rows=rows,
            fg=self.fg_color,
            bg=self.bg_color,
            path=self.path,
        ).replace("\n", "")

    @staticmethod
    def _kitty_cmd(chunk: "str" = "", **params: "Any") -> "str":
        param_str = ",".join(
            [f"{key}={value}" for key, value in params.items() if value is not None]
        )
        cmd = f"\x1b_G{param_str}"
        if chunk:
            cmd += f";{chunk}"
        cmd += "\x1b\\"
        return cmd

    def load(self, rows: "int", cols: "int") -> "None":
        """Sends the graphic to the terminal without displaying it."""
        global _kitty_image_count

        data = self.convert_data(rows=rows, cols=cols)
        self.kitty_image_id = _kitty_image_count
        _kitty_image_count += 1

        while data:
            chunk, data = data[:4096], data[4096:]
            cmd = self._kitty_cmd(
                chunk=chunk,
                a="t",  # We are sending an image without displaying it
                t="d",  # Transferring the image directly
                i=self.kitty_image_id,  # Send a unique image number, wait for an image id
                # I=self.kitty_image_number,  # Send a unique image number, wait for an image id
                p=1,  # Placement ID
                q=2,  # No chatback
                f=100,  # Sending a PNG image
                C=1,  # Do not move the cursor
                m=1 if data else 0,  # Data will be chunked
            )
            self.app.output.write_raw(tmuxify(cmd))
        self.app.output.flush()
        self.loaded = True

    def hide(self) -> "None":
        """Hides the graphic from show without deleting it."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(
                tmuxify(
                    self._kitty_cmd(
                        a="d",
                        d="i",
                        i=self.kitty_image_id,
                        q=1,
                    )
                )
            )
            self.app.output.flush()

    def delete(self) -> "None":
        """Deletes the graphic from the terminal."""
        if self.kitty_image_id > 0:
            self.app.output.write_raw(
                tmuxify(
                    self._kitty_cmd(
                        a="D",
                        d="I",
                        i=self.kitty_image_id,
                        q=2,
                    )
                )
            )
            self.app.output.flush()
            self.loaded = False

    def get_rendered_lines(
        self, width: "int", height: "int"
    ) -> "list[StyleAndTextTuples]":
        """Get rendered lines from the cache, or generate them."""
        # TODO - wezterm does not scale kitty graphics, so we might want to resize
        # images at this point rather than just loading them once
        if not self.loaded:
            self.load(cols=width, rows=height)

        def render_lines() -> "list[StyleAndTextTuples]":
            """Renders the lines to display in the control."""
            cmd = self._kitty_cmd(
                a="p",  # Display a previously transmitted image
                i=self.kitty_image_id,
                p=1,  # Placement ID
                m=0,  # No batches remaining
                q=2,  # No backchat
                c=width,
                r=height,
                C=1,  # 1 = Do move the cursor
                z=-(2**30) - 1,
            )
            return list(
                split_lines(
                    to_formatted_text(
                        [
                            # Move cursor down and acoss by image height and width
                            ("", "\n".join((height) * [" " * (width)])),
                            # Save position, then move back
                            (
                                "[ZeroWidthEscape]",
                                f"\x1b[s\x1b[{height-1}A\x1b[{width}D",
                            ),
                            # Place the image without moving cursor
                            ("[ZeroWidthEscape]", tmuxify(cmd)),
                            # Restore the last known cursor position (at the bottom)
                            ("[ZeroWidthEscape]", "\x1b[u"),
                        ]
                    )
                )
            )

        key = (width, self.app.term_info.cell_size_px)
        return self._format_cache.get(key, render_lines)

    def reset(self) -> "None":
        """Hide and delete the kitty graphic from the terminal."""
        self.hide()
        self.delete()
        super().reset()

    def close(self) -> "None":
        """Remove the displayed object entirely."""
        super().close()
        if not self.app.leave_graphics():
            self.delete()


class GraphicWindow(Window):
    """A window responsible for displaying terminal graphics content.

    The content is displays floating on top of a target window.

    The graphic will be displayed if:
    - a completion menu is not being shown
    - a dialog is not being shown
    - a menu is not being shown
    - the output it attached to is fully in view
    """

    content: "DisplayControl"

    def __init__(
        self,
        content: "GraphicControl",
        target_window: "Window",
        filter: "FilterOrBool",
        *args: "Any",
        **kwargs: "Any",
    ):
        """Initiates a new :py:class:`GraphicWindow` object.

        Args:
            content: A control which generates the graphical content to display
            target_window: The window this graphic should position itself over
            filter: A filter which determines if the graphic should be shown
            args: Positional arguments for :py:method:`Window.__init__`
            kwargs: Key-word arguments for :py:method:`Window.__init__`
        """
        super().__init__(*args, **kwargs)
        self.target_window = target_window
        self.content = content
        self.filter = to_filter(filter)

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Draws the graphic window's contents to the screen if required."""
        filter_value = self.filter()
        target_wp = screen.visible_windows_to_write_positions.get(self.target_window)
        if filter_value and target_wp and self.target_window.render_info is not None:
            # NOTE - using `render_info` here means `rendered_height` is potentially one
            # render cycle behind the `write_position`'s height. This only really
            # matters when the terminal is resized
            rendered_height = self.target_window.render_info.window_height
            # Only draw if the target window is fully visible
            if target_wp.height >= rendered_height:
                cpos = screen.get_menu_position(self.target_window)
                new_write_position = WritePosition(
                    xpos=cpos.x,
                    ypos=cpos.y,
                    width=target_wp.width,
                    height=target_wp.height,
                )

                super().write_to_screen(
                    screen,
                    MouseHandlers(),  # Do not let the float add mouse events
                    new_write_position,
                    # Force renderer refreshes by constantly changing the style
                    f"{parent_style} class:render-{get_app().render_counter}",
                    erase_bg=True,
                    z_index=z_index,
                )
                return
        # Otherwise hide the content (required for kitty graphics)
        if not filter_value or not get_app().leave_graphics():
            self.content.hide()

    def _fill_bg(
        self, screen: Screen, write_position: WritePosition, erase_bg: bool
    ) -> None:
        """Erase/fill the background."""
        char: Optional[str]
        if callable(self.char):
            char = self.char()
        else:
            char = self.char
        if erase_bg or char:
            wp = write_position
            char_obj = Char(char or " ", f"class:render-{get_app().render_counter}")
            for y in range(wp.ypos, wp.ypos + wp.height):
                row = screen.data_buffer[y]
                for x in range(wp.xpos, wp.xpos + wp.width):
                    row[x] = char_obj


class GraphicFloat(Float):
    """A :py:class:`Float` which displays a graphic."""

    def __init__(
        self,
        target_window: "Window",
        data: "Any",
        format_: "str",
        path: "Optional[UPath]" = None,
        fg_color: "Optional[str]" = None,
        bg_color: "Optional[str]" = None,
        sizing_func: "Optional[Callable[[], tuple[int, float]]]" = None,
        filter: "FilterOrBool" = True,
    ) -> "None":
        """Create a new instance.

        Args:
            data: The graphical data to be displayed
            format_: The format of the graphical data
            path: The path to the data's original location
            target_window: The window above which the graphic should be displayed
            fg_color: The graphic's foreground color
            bg_color: The graphic's background color
            sizing_func: A callable which returns in the graphic's width in terminal
                cells and its aspect ratio
            filter: A filter which is used to hide and show the graphic
        """
        self.GraphicControl: "Optional[Type[GraphicControl]]" = None
        self.control = None

        app = get_app()
        term_info = app.term_info
        if term_info.kitty_graphics_status.value and find_route(format_, "base64-png"):
            self.GraphicControl = KittyGraphicControl
        elif term_info.iterm_graphics_status.value and find_route(
            format_, "base64-png"
        ):
            self.GraphicControl = ItermGraphicControl
        elif term_info.sixel_graphics_status.value and find_route(format_, "sixel"):
            self.GraphicControl = SixelGraphicControl

        if self.GraphicControl:
            self.control = self.GraphicControl(
                data,
                format_=format_,
                path=path,
                fg_color=fg_color,
                bg_color=bg_color,
                sizing_func=sizing_func,
            )
            weak_self_ref = weakref.ref(self)
            super().__init__(
                content=GraphicWindow(
                    target_window=target_window,
                    content=self.control,
                    filter=to_filter(filter)
                    & (Condition(lambda: weak_self_ref() in get_app().graphics)),
                ),
                left=0,
                top=0,
            )
            # Hide the graphic if the float is deleted
            weakref.finalize(self, self.control.close)

    @property
    def data(self) -> "Any":
        """Return the graphic's current data."""
        return self._data

    @data.setter
    def data(self, value: "Any") -> "None":
        self._data = value
        if self.control is not None:
            self.control.data = value
            self.control.reset()


class Display:
    """Rich output displays.

    A container for displaying rich output data.

    """

    def __init__(
        self,
        data: "Any",
        format_: "str",
        path: "Optional[UPath]" = None,
        fg_color: "Optional[str]" = None,
        bg_color: "Optional[str]" = None,
        height: "AnyDimension" = None,
        width: "AnyDimension" = None,
        px: "int" = None,
        py: "int" = None,
        focusable: "FilterOrBool" = False,
        focus_on_click: "FilterOrBool" = False,
        wrap_lines: "FilterOrBool" = False,
        always_hide_cursor: "FilterOrBool" = True,
        scrollbar: "FilterOrBool" = True,
        scrollbar_autohide: "FilterOrBool" = True,
        dont_extend_height: "FilterOrBool" = True,
        style: "Union[str, Callable[[], str]]" = "",
    ) -> "None":
        """Instantiate an Output container object.

        Args:
            data: Raw cell output data
            format_: The conversion format of the data to render
            path: The path to the data's original location
            fg_color: The foreground colour to use when renderin this output
            bg_color: The background colour to use when renderin this output
            height: The height of the output
            width: The width of the output
            px: The pixel width of the data if known
            py: The pixel height of the data if known
            focusable: If the output should be focusable
            focus_on_click: If the output should become focused when clicked
            wrap_lines: If the output's lines should be wrapped
            always_hide_cursor: When true, the cursor is never shown
            scrollbar: Whether to show a scrollbar
            scrollbar_autohide: Whether to automatically hide the scrollbar
            dont_extend_height: Whether the window should fill the available height
            style: The style to apply to the output

        """
        self.style = style

        # Get data pixel dimensions
        self._px = px
        self._py = py

        sizing_func = self.make_sizing_func(
            data=data,
            format_=format_,
            fg=fg_color,
            bg=bg_color,
        )

        self.control = FormattedTextDisplayControl(
            data,
            format_=format_,
            path=path,
            fg_color=fg_color,
            bg_color=bg_color,
            sizing_func=sizing_func,
            focusable=focusable,
            focus_on_click=focus_on_click,
        )
        self.window = DisplayWindow(
            content=self.control,
            height=height,
            width=width,
            right_margins=[
                ConditionalMargin(
                    ScrollbarMargin(autohide=scrollbar_autohide),
                    filter=to_filter(scrollbar),
                )
            ],
            wrap_lines=wrap_lines,
            always_hide_cursor=always_hide_cursor,
            dont_extend_height=dont_extend_height,
            style=self.style,
            char=" ",
        )

        # Add graphic
        app = get_app()
        self.graphic_float = GraphicFloat(
            target_window=self.window,
            data=data,
            format_=format_,
            path=path,
            fg_color=fg_color,
            bg_color=bg_color,
            sizing_func=sizing_func,
            filter=~has_completions & ~has_dialog & ~has_menus,
        )
        if self.graphic_float.GraphicControl is not None:
            app.graphics.add(self.graphic_float)

    @property
    def data(self) -> "Any":
        """Return the display's current data."""
        return self.control.data

    @data.setter
    def data(self, value: "Any") -> "None":
        self.control.data = value
        self.graphic_float.data = value

    def make_sizing_func(
        self, data: "Any", format_: "str", fg: "Optional[str]", bg: "Optional[str]"
    ) -> "Callable[[], tuple[int, float]]":
        """Create a function to recalculate the data's dimensions in terminal cells."""
        px, py = self.px, self.py
        if px is None or py is None:
            px, py = data_pixel_size(data, format_, fg=fg, bg=bg)
        return partial(pixels_to_cell_size, px, py)

    @property
    def px(self) -> "Optional[int]":
        """Return the displayed data's pixel widget."""
        return self._px

    @px.setter
    def px(self, value: "Optional[int]") -> "None":
        self._px = value
        self.update_sizing()

    @property
    def py(self) -> "Optional[int]":
        """Return the displayed data's pixel height."""
        return self._py

    @py.setter
    def py(self, value: "Optional[int]") -> "None":
        self._py = value
        self.update_sizing()

    def update_sizing(self) -> "None":
        """Create a sizing function when the data's pixel size changes."""
        sizing_func = self.make_sizing_func(
            data=self.control.data,
            format_=self.control.format_,
            fg=self.control.fg_color,
            bg=self.control.bg_color,
        )
        self.control.sizing_func = sizing_func
        self.control.reset()
        if self.graphic_float.control is not None:
            self.graphic_float.control.sizing_func = sizing_func
            self.graphic_float.control.reset()

    def __pt_container__(self) -> "AnyContainer":
        """Return the content of this output."""
        return self.window

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_left() -> "None":
        """Scroll the display up one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_left()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_right() -> "None":
        """Scroll the display down one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_right()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_up() -> "None":
        """Scroll the display up one line."""
        get_app().layout.current_window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_down() -> "None":
        """Scroll the display down one line."""
        get_app().layout.current_window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_up_display() -> "None":
        """Scroll the display up one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_down_display() -> "None":
        """Scroll the display down one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_start_of_display() -> "None":
        """Scroll the display to the top."""
        from euporie.core.widgets.display import DisplayControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, DisplayControl):
            current_control.cursor_position = Point(0, 0)

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_end_of_display() -> "None":
        """Scroll the display down one page."""
        from euporie.core.widgets.display import DisplayControl

        layout = get_app().layout
        current_control = layout.current_control
        window = layout.current_window
        if (
            isinstance(current_control, DisplayControl)
            and window.render_info is not None
        ):
            current_control.cursor_position = Point(
                0, window.render_info.ui_content.line_count - 1
            )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.widgets.display.Display": {
                "scroll-display-left": "left",
                "scroll-display-right": "right",
                "scroll-display-up": ["up", "k"],
                "scroll-display-down": ["down", "j"],
                "page-up-display": "pageup",
                "page-down-display": "pagedown",
                "go-to-start-of-display": "home",
                "go-to-end-of-display": "end",
            }
        }
    )



"""Contains the main class for a notebook file."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, cast

import nbformat
from prompt_toolkit.application.run_in_terminal import in_terminal
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.filters.app import (
    buffer_has_focus,
    has_completions,
    has_focus,
    has_selection,
    in_paste_mode,
    renderer_height_is_known,
)
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout

from euporie.core.commands import add_cmd, get_cmd
from euporie.core.config import add_setting
from euporie.core.filters import (
    at_end_of_buffer,
    buffer_is_code,
    buffer_is_empty,
    kernel_tab_has_focus,
)
from euporie.core.format import format_code
from euporie.core.kernel import MsgCallbacks
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.style import KERNEL_STATUS_REPR
from euporie.core.tabs.base import KernelTab
from euporie.core.validation import KernelValidator
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.inputs import KernelInput, StdInput
from euporie.core.widgets.page import PrintingContainer
from euporie.core.widgets.pager import PagerState

if TYPE_CHECKING:
    from typing import Any, Callable, Optional, Sequence

    from prompt_toolkit.application.application import Application
    from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import Float
    from upath import UPath

    from euporie.core.app import BaseApp

log = logging.getLogger(__name__)


class Console(KernelTab):
    """Interactive console.

    An interactive console which connects to a Jupyter kernel.

    """

    def __init__(
        self,
        app: "BaseApp",
        path: "Optional[UPath]" = None,
        use_kernel_history: "bool" = True,
    ) -> "None":
        """Create a new :py:class:`KernelNotebook` instance.

        Args:
            app: The euporie application the console tab belongs to
            path: A file path to open (not used currently)
            use_kernel_history: If :const:`True`, history will be loaded from the kernel
        """
        # Kernel setup
        self._metadata = {}
        self.kernel_name = app.config.kernel_name
        self.allow_stdin = True
        self.default_callbacks = MsgCallbacks(
            get_input=lambda prompt, password: self.stdin_box.get_input(
                prompt, password
            ),
            set_execution_count=partial(setattr, self, "execution_count"),
            add_output=self.new_output,
            clear_output=self.clear_output,
            # set_metadata=self.misc_callback,
            set_status=lambda status: self.app.invalidate(),
            set_kernel_info=self.set_kernel_info,
            done=self.complete,
        )
        self.kernel_tab = self

        super().__init__(app=app, path=path, use_kernel_history=use_kernel_history)

        self.lang_info: "dict[str, Any]" = {}
        self.execution_count = 0
        self.clear_outputs_on_output = False

        self.json = nbformat.v4.new_notebook()
        self.json["metadata"] = self._metadata

        self.container = self.load_container()

        self.kernel.start(cb=self.kernel_started, wait=False)

        self.app.before_render += self.render_outputs

    async def load_history(self) -> "None":
        """Load kernel history."""
        await super().load_history()
        # Re-run history load for the input-box
        self.input_box.buffer._load_history_task = None
        self.input_box.buffer.load_history_if_not_yet_loaded()

    def close(self, cb: "Optional[Callable]" = None) -> "None":
        """Close the console tab."""
        # Ensure any output no longer appears interactive
        self.live_output.style = "class:disabled"
        # Unregister output renderer
        self.app.before_render -= self.render_outputs
        super().close(cb)

    def clear_output(self, wait: "bool" = False) -> "None":
        """Remove the last output, optionally when new output is generated."""
        if wait:
            self.clear_outputs_on_output = True
        else:
            self.live_output.reset()

    def validate_input(self, code: "str") -> "bool":
        """Determine if the entered code is ready to run."""
        assert self.kernel is not None
        completeness_status = self.kernel.is_complete(code, wait=True).get(
            "status", "unknown"
        )
        if (
            not code.strip()
            or completeness_status == "incomplete"
            or (completeness_status == "unknown" and code[-2:] != "\n\n")
        ):
            return False
        else:
            return True

    def run(self, buffer: "Optional[Buffer]" = None) -> "None":
        """Run the code in the input box."""
        if buffer is None:
            buffer = self.input_box.buffer
        # Auto-reformat code
        if self.app.config.autoformat:
            self.reformat()
        # Get the code to run
        text = buffer.text
        # Remove any selections from input
        self.input_box.buffer.selection_state = None
        # Disable existing output
        self.live_output.style = "class:disabled"
        # Re-render the app and move to below the current output
        self.app.draw()
        # Prevent displayed graphics on terminal being cleaned up (bit of a hack)
        self.app.graphics.clear()
        # Run the previous entry
        if self.kernel.status == "starting":
            self.kernel_queue.append(partial(self.kernel.run, text, wait=False))
        else:
            self.kernel.run(text, wait=False)
        # Increment this for display purposes until we get the response from the kernel
        self.execution_count += 1
        # Reset the input & output
        buffer.reset(append_to_history=True)
        # Remove any live outputs and disable mouse support
        self.live_output.reset()
        if self.app.config.mouse_support is None:
            self.app.need_mouse_support = False
        # Record the input as a cell in the json
        self.json["cells"].append(
            nbformat.v4.new_code_cell(source=text, execution_count=self.execution_count)
        )
        if (
            self.app.config.max_stored_outputs
            and len(self.json["cells"]) > self.app.config.max_stored_outputs
        ):
            del self.json["cells"][0]

    def new_output(self, output_json: "dict[str, Any]") -> "None":
        """Print the previous output and replace it with the new one."""
        # Clear the output if we were previously asked to
        if self.clear_outputs_on_output:
            self.clear_outputs_on_output = False
            # Clear the screen
            get_cmd("clear-screen").run()

        # Add to record
        if self.json["cells"]:
            self.json["cells"][-1]["outputs"].append(output_json)

        # Set output
        if "application/vnd.jupyter.widget-view+json" in output_json.get("data", {}):
            # Use a live output to display widgets
            self.live_output.add_output(output_json)
            # Enable mouse support if we have a live output
            if self.app.config.mouse_support is None:
                self.app.need_mouse_support = True
        else:
            # Queue the output json
            self.output.add_output(output_json)
            # Invalidate the app so the output get printed
            self.app.invalidate()

    def render_outputs(self, app: "Application[Any]") -> "None":
        """Request that any unrendered outputs be rendered."""
        if self.output.json:
            self.app.create_background_task(self.async_render_outputs())

    async def async_render_outputs(self) -> "None":
        """Render any unrendered outputs above the application."""
        if self.output.json:
            # Run the output app in the terminal
            async with in_terminal():
                self.app.renderer.render(self.app, self.output_layout, is_done=True)
            # Remove the outputs so they do not get rendered again
            self.output.reset()

    def reset(self) -> "None":
        """Reset the state of the tab."""
        self.live_output.reset()
        if self.app.config.mouse_support is None:
            self.app.need_mouse_support = False

    def complete(self, content: "dict" = None) -> "None":
        """Re-render any changes."""
        self.app.invalidate()

    def prompt(
        self, text: "str", offset: "int" = 0, show_busy: "bool" = False
    ) -> "StyleAndTextTuples":
        """Determine what should be displayed in the prompt of the cell."""
        prompt = str(self.execution_count + offset)
        if show_busy and self.kernel.status in ("busy", "queued"):
            prompt = "*".center(len(prompt))
        ft: "StyleAndTextTuples" = [
            ("", f"{text}["),
            ("class:count", prompt),
            ("", "]: "),
        ]
        return ft

    @property
    def language(self) -> "str":
        """The language of the current kernel."""
        return self.lang_info.get(
            "name", self.lang_info.get("pygments_lexer", "python")
        )

    def lang_file_ext(self) -> "str":
        """Return the file extension for scripts in the notebook's language."""
        return self.lang_info.get("file_extension", ".py")

    def load_container(self) -> "HSplit":
        """Builds the main application layout."""
        # Output area

        self.output = CellOutputArea([], parent=self)

        @Condition
        def first_output() -> "bool":
            """Check if the current outputs contain the first output."""
            if self.output.json:
                for output in self.json["cells"][-1].get("outputs", []):
                    if output in self.live_output.json:
                        continue
                    return output == self.output.json[0]
            return False

        output_prompt = Window(
            FormattedTextControl(partial(self.prompt, "Out", show_busy=False)),
            dont_extend_width=True,
            style="class:cell.output.prompt",
            height=1,
        )
        output_margin = Window(
            char=" ", width=lambda: len(str(self.execution_count)) + 7
        )

        output_area = PrintingContainer(
            [
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True), filter=first_output
                ),
                VSplit(
                    [
                        ConditionalContainer(output_prompt, filter=first_output),
                        ConditionalContainer(output_margin, filter=~first_output),
                        self.output,
                    ]
                ),
            ],
        )
        # We need to ensure the output layout also has the application's floats so that
        # graphics are displayed
        self.output_layout = Layout(
            FloatContainer(
                output_area,
                floats=cast("list[Float]", self.app.graphics),
            )
        )

        # Live output area

        self.live_output = CellOutputArea([], parent=self)

        # Input area

        def on_cursor_position_changed(buf: "Buffer") -> "None":
            """Respond to cursor movements."""
            # Update contextual help
            if self.app.config.autoinspect and buf.name == "code":
                self.inspect()
            elif (pager := self.app.pager) is not None and pager.visible():
                pager.hide()

        input_kb = KeyBindings()

        @Condition
        def empty() -> "bool":
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            text = buffer.text
            return not text.strip()

        @Condition
        def not_invalid() -> "bool":
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            return buffer.validation_state != ValidationState.INVALID

        @input_kb.add(
            "enter",
            filter=has_focus("code")
            & ~empty
            & not_invalid
            & at_end_of_buffer
            & ~has_completions,
        )
        async def on_enter(event: "KeyPressEvent") -> "NotImplementedOrNone":
            """Accept input if the input is valid, otherwise insert a return."""
            buffer = event.current_buffer
            # Accept the buffer if there are 2 blank lines
            accept = buffer.text[-2:] == "\n\n"
            # Also accept if the input is valid
            if not accept:
                accept = buffer.validate(set_cursor=False)
            if accept:
                if buffer.accept_handler:
                    keep_text = buffer.accept_handler(buffer)
                else:
                    keep_text = False
                # buffer.append_to_history()
                if not keep_text:
                    buffer.reset()
                return NotImplemented

            # Process the input as a regular :kbd:`enter` key-press
            event.key_processor.feed(event.key_sequence[0], first=True)
            # Prevent the app getting invalidated
            return None

        @input_kb.add("s-enter")
        def _newline(event: "KeyPressEvent") -> "None":
            """Force new line on Shift-Enter."""
            event.current_buffer.newline(copy_margin=not in_paste_mode())

        input_prompt = Window(
            FormattedTextControl(partial(self.prompt, "In ", offset=1)),
            dont_extend_width=True,
            style="class:cell.input.prompt",
            height=1,
        )

        self.input_box = KernelInput(
            kernel_tab=self,
            accept_handler=self.run,
            on_cursor_position_changed=on_cursor_position_changed,
            validator=KernelValidator(self.kernel),
            # validate_while_typing=False,
            enable_history_search=True,
            key_bindings=input_kb,
        )
        self.input_box.buffer.name = "code"

        self.app.focused_element = self.input_box.buffer

        self.stdin_box = StdInput(self)

        self.input_layout = HSplit(
            [
                ConditionalContainer(
                    HSplit(
                        [
                            Window(height=1, dont_extend_height=True),
                            VSplit([output_margin, self.live_output]),
                        ]
                    ),
                    filter=Condition(lambda: bool(self.live_output.json)),
                ),
                # StdIn
                self.stdin_box,
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True),
                    filter=self.stdin_box.visible,
                ),
                # Spacing
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True),
                    filter=Condition(lambda: self.execution_count > 0)
                    & (
                        (
                            renderer_height_is_known
                            & Condition(lambda: self.app.renderer.rows_above_layout > 0)
                        )
                        | ~renderer_height_is_known
                    ),
                ),
                # Input
                ConditionalContainer(
                    VSplit(
                        [
                            input_prompt,
                            self.input_box,
                        ],
                    ),
                    filter=~self.stdin_box.visible,
                ),
            ],
            key_bindings=load_registered_bindings(
                "euporie.console.tabs.console.Console"
            ),
        )

        return self.input_layout

    def accept_stdin(self, buf: "Buffer") -> "bool":
        """Accept the user's input."""
        return True

    def interrupt_kernel(self) -> "None":
        """Interrupt the current `Notebook`'s kernel."""
        assert self.kernel is not None
        self.kernel.interrupt()

    @property
    def kernel_display_name(self) -> "str":
        """Return the display name of the kernel defined in the notebook JSON."""
        if self.kernel and self.kernel.km.kernel_spec:
            return self.kernel.km.kernel_spec.display_name
        return self.kernel_name

    def set_kernel_info(self, info: "dict") -> "None":
        """Receives and processes kernel metadata."""
        self.lang_info = info.get("language_info", {})

    def refresh(self, now: "bool" = True) -> "None":
        """Request the output is refreshed (does nothing)."""
        pass

    def statusbar_fields(
        self,
    ) -> "tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]":
        """Generates the formatted text for the statusbar."""
        assert self.kernel is not None
        return (
            [],
            [
                [
                    (
                        "",
                        self.kernel_display_name
                        # , self._statusbar_kernel_handeler)],
                    )
                ],
                KERNEL_STATUS_REPR.get(self.kernel.status, "."),
            ],
        )

    def reformat(self) -> "None":
        """Reformats the input."""
        self.input_box.text = format_code(self.input_box.text, self.app.config)

    def inspect(self) -> "None":
        """Get contextual help for the current cursor position in the current cell."""
        code = self.input_box.text
        cursor_pos = self.input_box.buffer.cursor_position

        assert self.app.pager is not None

        if self.app.pager.visible() and self.app.pager.state is not None:
            if (
                self.app.pager.state.code == code
                and self.app.pager.state.cursor_pos == cursor_pos
            ):
                self.app.pager.focus()
                return

        def _cb(response: "dict") -> "None":
            assert self.app.pager is not None
            prev_state = self.app.pager.state
            new_state = PagerState(
                code=code,
                cursor_pos=cursor_pos,
                response=response,
            )
            if prev_state != new_state:
                self.app.pager.state = new_state
                self.app.invalidate()

        assert self.kernel is not None
        self.kernel.inspect(
            code=code,
            cursor_pos=cursor_pos,
            callback=_cb,
        )

    def save(self, path: "UPath" = None, cb: "Optional[Callable]" = None) -> "None":
        """Save the console as a notebook."""
        from euporie.core.tabs.notebook import BaseNotebook

        BaseNotebook.save(cast("BaseNotebook", self), path)

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _accept_input() -> "None":
        """Accept the current console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        if buffer:
            buffer.validate_and_handle()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection & ~buffer_is_empty,
    )
    def _clear_input() -> "None":
        """Clear the console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        buffer.reset()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus,
    )
    def _run_input() -> "None":
        """Run the console input."""
        from euporie.console.app import get_app

        console = get_app().tab
        assert isinstance(console, Console)
        console.run()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection,
    )
    def _show_contextual_help() -> "None":
        """Displays contextual help."""
        from euporie.console.app import get_app

        console = get_app().tab
        assert isinstance(console, Console)
        console.inspect()

    @staticmethod
    @add_cmd(
        name="cc-interrupt-kernel",
        hidden=True,
        filter=buffer_is_code & buffer_is_empty,
    )
    @add_cmd(
        filter=kernel_tab_has_focus,
    )
    def _interrupt_kernel() -> "None":
        """Interrupt the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.interrupt_kernel()

    @staticmethod
    @add_cmd(
        filter=kernel_tab_has_focus,
    )
    def _restart_kernel() -> "None":
        """Restart the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.restart_kernel()

    # ################################### Settings ####################################

    add_setting(
        name="max_stored_outputs",
        flags=["--max-stored-outputs"],
        type_=int,
        help_="The number of inputs / outputs to store in an in-memory notebook",
        default=100,
        schema={
            "minimum": 0,
        },
        description="""
            Defines the maximum number of executed "cells" to store in case the console
            session is saved to a file or converted into a notebook.
        """,
    )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.console.tabs.console.Console": {
                "cc-interrupt-kernel": "c-c",
                "show-contextual-help": "s-tab",
            },
            "euporie.console.app.ConsoleApp": {
                "clear-input": "c-c",
                "run-input": ["c-enter", "c-e"],
            },
        }
    )



"""Decorative widgets."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import has_focus, to_filter
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    DynamicContainer,
    HSplit,
    VSplit,
    Window,
    to_container,
)
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.layout.mouse_handlers import MouseHandlers
from prompt_toolkit.layout.screen import Char, Screen, WritePosition

from euporie.core.border import BorderVisibility, Thin

if TYPE_CHECKING:
    from typing import Callable, Optional, Union

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.border import GridStyle

log = logging.getLogger(__name__)


class Line(Container):
    """Draws a horizontal or vertical line."""

    def __init__(
        self,
        char: "Optional[str]" = None,
        width: "Optional[int]" = None,
        height: "Optional[int]" = None,
        collapse: "bool" = False,
        style: "str" = "class:grid-line",
    ) -> "None":
        """Initializes a grid line.

        Args:
            char: The character to draw. If unset, the relevant character from
                :py:class:`euporie.core.box.grid` is used
            width: The length of the line. If specified, the line will be horizontal
            height: The height of the line. If specified, the line will be vertical
            collapse: Whether to hide the line when there is not enough space
            style: Style to apply to the line

        Raises:
            ValueError: If both width and height are specified. A line must only have a
                single dimension.

        """
        if width and height:
            raise ValueError("Only one of `width` or `height` must be set")
        self.width = width
        self.height = height
        if char is None:
            char = Thin.grid.VERTICAL if width else Thin.grid.HORIZONTAL
        self.char = Char(char, style)
        self.collapse = collapse

    def reset(self) -> "None":
        """Resets the state of the line. Does nothing."""

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Return the preferred width of the line."""
        return Dimension(min=int(not self.collapse), max=self.width)

    def preferred_height(
        self, width: "int", max_available_height: "int"
    ) -> "Dimension":
        """Return the preferred height of the line."""
        return Dimension(min=int(not self.collapse), max=self.height)

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Draws a continuous line in the ``write_position`` area.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        ypos = write_position.ypos
        xpos = write_position.xpos

        for y in range(ypos, ypos + write_position.height):
            row = screen.data_buffer[y]
            for x in range(xpos, xpos + write_position.width):
                row[x] = self.char

    def get_children(self) -> "list":
        """Return an empty list of the container's children."""
        return []


class Pattern(Container):
    """Fill an area with a repeating background pattern."""

    def __init__(
        self,
        char: "Union[str, Callable[[], str]]",
        pattern: "Union[int, Callable[[], int]]" = 1,
    ) -> "None":
        """Initialize the :class:`Pattern`."""
        self.bg = Char(" ", "class:pattern")
        self.char = char
        self.pattern = pattern

    def reset(self) -> "None":
        """Resets the pattern. Does nothing."""
        pass

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Return an empty dimension (expand to available width)."""
        return Dimension()

    def preferred_height(
        self, width: "int", max_available_height: "int"
    ) -> "Dimension":
        """Return an empty dimension (expand to available height)."""
        return Dimension()

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Fill the whole area of write_position with a pattern.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        bg = self.bg
        if callable(self.pattern):
            pattern = self.pattern()
        else:
            pattern = self.pattern
        if callable(self.char):
            char = Char(self.char(), "class:pattern")
        else:
            char = Char(self.char, "class:pattern")

        ypos = write_position.ypos
        xpos = write_position.xpos

        for y in range(ypos, ypos + write_position.height):
            row = screen.data_buffer[y]
            for x in range(xpos, xpos + write_position.width):
                if (
                    (pattern == 1)
                    or (pattern == 2 and (x + y) % 2 == 0)
                    or (pattern == 3 and (x + 2 * y) % 4 == 0)
                    or (pattern == 4 and (x + y) % 3 == 0)
                    or (pattern == 5 and ((x + y % 2 * 3) % 6) % 4 == 0)
                ):
                    row[x] = char
                else:
                    row[x] = bg

    def get_children(self) -> "list":
        """Return an empty list of the container's children."""
        return []


class Border:
    """Draw a border around any container."""

    def __init__(
        self,
        body: "AnyContainer",
        border: "Optional[GridStyle]" = Thin.grid,
        style: "Union[str, Callable[[], str]]" = "class:frame.border",
        show_borders: "Optional[BorderVisibility]" = None,
    ) -> "None":
        """Create a new border widget which wraps another container.

        Args:
            body: The container to surround with a border
            border: The grid style to use
            style: The style to apply to the border
            show_borders: Which of the four borders should be displayed

        """
        self.body = body
        self.style = style

        if show_borders:
            show_borders = BorderVisibility(*show_borders)
        else:
            show_borders = BorderVisibility(True, True, True, True)
        border_top = to_filter(show_borders.top)
        border_right = to_filter(show_borders.right)
        border_bottom = to_filter(show_borders.bottom)
        border_left = to_filter(show_borders.left)

        self.container: "AnyContainer"
        if border is not None and any(show_borders):
            self.container = HSplit(
                [
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_LEFT,
                                        style=self.add_style("class:left,top"),
                                    ),
                                    filter=border_top & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.TOP_MID,
                                        style=self.add_style("class:top"),
                                    ),
                                    filter=border_top,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_RIGHT,
                                        style=self.add_style("class:right,top"),
                                    ),
                                    filter=border_top & border_right,
                                ),
                            ],
                            height=1,
                        ),
                        filter=border_top,
                    ),
                    VSplit(
                        [
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_LEFT,
                                    style=self.add_style("class:left"),
                                ),
                                filter=border_left,
                            ),
                            DynamicContainer(lambda: self.body),
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_RIGHT,
                                    style=self.add_style("class:right"),
                                ),
                                filter=border_right,
                            )
                            # Padding is required to make sure that if the content is
                            # too small, the right frame border is still aligned.
                        ],
                        padding=0,
                    ),
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_LEFT,
                                        style=self.add_style("class:left,bottom"),
                                    ),
                                    filter=border_bottom & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.BOTTOM_MID,
                                        style=self.add_style("class:bottom"),
                                    ),
                                    filter=border_bottom,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_RIGHT,
                                        style=self.add_style("class:right,bottom"),
                                    ),
                                    filter=border_bottom & border_right,
                                ),
                            ],
                            # specifying height here will increase the rendering speed.
                            height=1,
                        ),
                        filter=border_bottom,
                    ),
                ],
            )
        else:
            self.container = body

    def add_style(self, extra: "str") -> "Callable[[], str]":
        """Return a function which adds a style string to the border style."""

        def _style() -> "str":
            if callable(self.style):
                return f"{self.style()} {extra}"
            else:
                return f"{self.style} {extra}"

        return _style

    def __pt_container__(self) -> "AnyContainer":
        """Return the border widget's container."""
        return self.container


class FocusedStyle(Container):
    """Applies a style to child containers when focused."""

    def __init__(
        self,
        body: "AnyContainer",
        style: "Union[str, Callable[[], str]]" = "class:focused",
    ) -> "None":
        """Create a new instance of the widget.

        Args:
            body: The container to act on
            style: The style to apply
        """
        self.body = body
        self.style = style
        self.has_focus = has_focus(self.body)

    def reset(self) -> "None":
        """Reset the wrapped container."""
        to_container(self.body).reset()

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Return the wrapped container's preferred width."""
        return to_container(self.body).preferred_width(max_available_width)

    def preferred_height(
        self, width: "int", max_available_height: "int"
    ) -> "Dimension":
        """Return the wrapped container's preferred height."""
        return to_container(self.body).preferred_height(width, max_available_height)

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Draw the wrapped container with the additional style."""
        return to_container(self.body).write_to_screen(
            screen,
            mouse_handlers,
            write_position,
            f"{parent_style} {self.get_style()}",
            erase_bg,
            z_index,
        )

    def get_style(self) -> "str":
        """Determine the style to apply depending on the focus status."""
        if self.has_focus():
            style = self.style
            return style() if callable(style) else style
        else:
            return ""

    def get_children(self) -> "list[Container]":
        """Return the list of child :class:`.Container` objects."""
        return [to_container(self.body)]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['euporie',
 'euporie.console',
 'euporie.console.tabs',
 'euporie.core',
 'euporie.core.comm',
 'euporie.core.convert',
 'euporie.core.convert.formats',
 'euporie.core.formatted_text',
 'euporie.core.key_binding',
 'euporie.core.key_binding.bindings',
 'euporie.core.tabs',
 'euporie.core.widgets',
 'euporie.hub',
 'euporie.notebook',
 'euporie.notebook.tabs',
 'euporie.preview',
 'euporie.preview.tabs']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.0,<10.0',
 'Pygments>=2.11.2,<3.0.0',
 'aenum>=3.1,<4.0',
 'appdirs>=1.4,<2.0',
 'fastjsonschema>=2.15.3,<3.0.0',
 'flatlatex>=0.15,<0.16',
 'fsspec[http]>=2022.8.0,<2023.0.0',
 'imagesize>=1.3.0,<2.0.0',
 'jupyter-client>=7.1,<8.0',
 'linkify-it-py>=1,<2',
 'markdown-it-py[linkify]>=2.1.0,<3.0.0',
 'mdit-py-plugins>=0.3.0,<0.4.0',
 'nbformat>=5,<6',
 'prompt-toolkit>=3.0.27,<4.0.0',
 'pyperclip>=1.8,<2.0',
 'timg>=1.1,<2.0',
 'typing-extensions>=4.2.0,<5.0.0',
 'universal-pathlib>=0.0.19,<0.1.0']

extras_require = \
{'all': ['asyncssh>=2.10.1,<3.0.0',
         'black>=19.3b0',
         'isort>=5.10.1,<6.0.0',
         'ssort>=0.11.6,<0.12.0',
         'CairoSVG>=2.5,<3.0'],
 'formatters': ['black>=19.3b0',
                'isort>=5.10.1,<6.0.0',
                'ssort>=0.11.6,<0.12.0'],
 'html-mtable': ['mtable>=0.1,<0.2', 'html5lib>=1.1,<2.0'],
 'hub': ['asyncssh>=2.10.1,<3.0.0'],
 'images-img2unicode': ['img2unicode>=0.1a8,<0.2'],
 'latex-sympy': ['sympy>=1.9,<2.0', 'antlr4-python3-runtime>=4.9,<5.0'],
 'svg-cairosvg': ['CairoSVG>=2.5,<3.0']}

entry_points = \
{'console_scripts': ['euporie = euporie.core.__main__:main',
                     'euporie-console = euporie.console.__main__:main',
                     'euporie-hub = euporie.hub.__main__:main',
                     'euporie-notebook = euporie.notebook.__main__:main',
                     'euporie-preview = euporie.preview.__main__:main'],
 'euporie.apps': ['console = euporie.console.app:ConsoleApp',
                  'hub = euporie.hub.app:HubApp',
                  'launch = euporie.core.launch:CoreApp',
                  'notebook = euporie.notebook.app:NotebookApp',
                  'preview = euporie.preview.app:PreviewApp'],
 'pygments.lexers': ['argparse = euporie.core.pygments:ArgparseLexer'],
 'pygments.styles': ['euporie = euporie.core.pygments:EuporiePygmentsStyle']}

setup_kwargs = {
    'name': 'euporie',
    'version': '2.1.5',
    'description': 'Euporie is a suite of terminal applications for interacting with Jupyter kernels',
    'long_description': '|logo|\n\n.. |logo| image:: https://user-images.githubusercontent.com/12154190/160670889-c6fc4cd8-413d-49f0-b105-9c0e03117032.svg\n   :alt: <Logo>\n\n#######\neuporie\n#######\n\n|PyPI| |RTD| |PyVer| |License| |Binder| |Stars|\n\n.. content_start\n\n**Euporie is a terminal based interactive computing environment for Jupyter.**\n\nEuporie\'s apps allow you to interact with Jupyter kernels, and run Jupyter notebooks - entirely from the terminal.\n\n.. list-table::\n   :align: center\n   :widths: 25 25 25 25\n   :class: text-center\n\n   * - `Console <https://euporie.readthedocs.io/en/latest/apps/console.html>`_\n     - `Notebook <https://euporie.readthedocs.io/en/latest/apps/notebook.html>`_\n     - `Preview <https://euporie.readthedocs.io/en/latest/apps/preview.html>`_\n     - `Hub <https://euporie.readthedocs.io/en/latest/apps/hub.html>`_\n\n.. image:: https://user-images.githubusercontent.com/12154190/182201621-9c3e4b04-1b1e-46e3-b852-b43f16adfc7b.png\n   :target: https://user-images.githubusercontent.com/12154190/182201621-9c3e4b04-1b1e-46e3-b852-b43f16adfc7b.png\n\n`View more screenshots here <https://euporie.readthedocs.io/en/latest/pages/gallery.html>`_\n\n----\n\n*******\nInstall\n*******\n\nYou can install euporie with `pipx <https://pipxproject.github.io/>`_ (recommended) or ``pip``:\n\n.. code-block:: console\n\n   $ pipx install euporie\n   $ # OR\n   $ python -m pip install --user euporie\n\nYou can also try euporie online `here <https://mybinder.org/v2/gh/joouha/euporie-binder/HEAD?urlpath=%2Feuporie%2F>`_.\n\n********\nFeatures\n********\n\n* Edit and run notebooks in the terminal\n* Run code interactively in a console\n* Display images using terminal graphics (sixel / iterm / kitty)\n* Use Jupyter widgets interactively in the terminal\n* Render rich kernel output (markdown, tables, images, LaTeX, HTML, SVG, & PDF)\n* Tab-completion, line suggestions  and contextual help\n* Convert a console session to a notebook\n* Micro / Vim / Emacs style key-bindings\n\n*****\nUsage\n*****\n\n**Notebooks**\n\n   You can edit a notebook using ``euporie-notebook``, and passing the notebook\'s file path or URI as a command line argument:\n\n   .. code-block:: console\n\n      $ euporie-notebook notebook.ipynb\n\n   Alternatively, launch ``euporie-notebooks`` and open a notebook file by selecting "Open" from the file menu (``Ctrl+O``).\n\n**Console**\n\n   To connect to a Jupyter kernel and run code interactively in a console session, you can run\n\n   .. code-block:: console\n\n      $ euporie-console\n\n   (You can press ``Ctrl+C`` to open the command palette in ``euporie-console``).\n\n**Preview**\n\n   To preview a notebook to the terminal, use the ``euporie-preview`` subcommand:\n\n   .. code-block:: console\n\n      $ euporie-preview notebook.ipynb\n\n**Hub**\n\n   To run euporie hub, a multi-user SSH server for euporie apps, run:\n\n   .. code-block:: console\n\n      $ euporie-hub --port 8022 --host-keys=ssh_host_ed25519_key --client-keys=authorized_keys\n\n   where ``ssh_host_ed25519_key`` is the path to your host key file, and ``authorized_keys`` is a file containing SSH public keys allowed to connect.\n\n*************\nDocumentation\n*************\n\nView the online documentation at: `https://euporie.readthedocs.io/ <https://euporie.readthedocs.io/>`_\n\nThe code is available on GitHub at: `https://github.com/joouha/euporie <https://github.com/joouha/euporie>`_\n\n*************\nCompatibility\n*************\n\nEuporie requires Python 3.8 or later. It works on Linux, Windows and MacOS\n\n\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/euporie.svg\n    :target: https://pypi.python.org/project/euporie/\n    :alt: Latest Version\n\n.. |RTD| image:: https://readthedocs.org/projects/euporie/badge/\n    :target: https://euporie.readthedocs.io/en/latest/\n    :alt: Documentation\n\n.. |PyVer| image:: https://img.shields.io/pypi/pyversions/euporie\n    :target: https://pypi.python.org/project/euporie/\n    :alt: Supported Python versions\n\n.. |Binder| image:: https://mybinder.org/badge_logo.svg\n   :target: https://mybinder.org/v2/gh/joouha/euporie-binder/HEAD?urlpath=%2Feuporie%2F\n   :alt: Launch with Binder\n\n.. |License| image:: https://img.shields.io/github/license/joouha/euporie.svg\n    :target: https://github.com/joouha/euporie/blob/main/LICENSE\n    :alt: View license\n\n.. |Stars| image:: https://img.shields.io/github/stars/joouha/euporie\n    :target: https://github.com/joouha/euporie/stargazers\n    :alt: ⭐\n',
    'author': 'Josiah Outram Halstead',
    'author_email': 'josiah@halstead.email',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/joouha/euporie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Contains tab base classes."""

from __future__ import annotations

import logging
from abc import ABCMeta
from collections import deque
from typing import TYPE_CHECKING

from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.layout.containers import Window

from euporie.core.app import get_app
from euporie.core.comm.registry import open_comm
from euporie.core.commands import add_cmd
from euporie.core.completion import KernelCompleter
from euporie.core.config import add_setting
from euporie.core.filters import kernel_tab_has_focus, tab_has_focus
from euporie.core.history import KernelHistory
from euporie.core.kernel import Kernel, MsgCallbacks
from euporie.core.suggest import HistoryAutoSuggest

if TYPE_CHECKING:
    from typing import Any, Callable, Deque, Optional, Sequence

    from prompt_toolkit.auto_suggest import AutoSuggest
    from prompt_toolkit.completion.base import Completer
    from prompt_toolkit.formatted_text import AnyFormattedText
    from prompt_toolkit.history import History
    from prompt_toolkit.layout.containers import AnyContainer
    from upath import UPath

    from euporie.core.app import BaseApp
    from euporie.core.comm.base import Comm

log = logging.getLogger(__name__)


class Tab(metaclass=ABCMeta):
    """Base class for interface tabs."""

    container: "AnyContainer"

    def __init__(self, app: "BaseApp", path: "Optional[UPath]" = None):
        """Called when the tab is created."""
        self.app = app
        self.path = path
        self.app.container_statuses[self] = self.statusbar_fields
        self.container = Window()
        self.kernel_queue: "Deque[Callable]" = deque()

    def statusbar_fields(
        self,
    ) -> "tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]":
        """Returns a list of statusbar field values shown then this tab is active."""
        return ([], [])

    @property
    def title(self) -> "str":
        """Return the tab title."""
        return ""

    def reset(self) -> "None":
        """Reset the state of the tab."""

    def close(self, cb: "Optional[Callable]" = None) -> "None":
        """Function to close a tab with a callback.

        Args:
            cb: A function to call after the tab is closed.

        """
        if self in self.app.container_statuses:
            del self.app.container_statuses[self]
        if callable(cb):
            cb()

    def focus(self) -> "None":
        """Focuses the tab (or make it visible)."""
        self.app.focus_tab(self)

    def save(self, path: "UPath" = None, cb: "Optional[Callable]" = None) -> "None":
        """Save the current notebook."""
        raise NotImplementedError

    def __pt_container__(self) -> "AnyContainer":
        """Return the main container object."""
        return self.container

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=tab_has_focus, title="Reset the current tab")
    def _reset_tab() -> "None":
        """Reset the current tab, reloading contents from source."""
        if (tab := get_app().tab) is not None:
            tab.reset()


class KernelTab(Tab, metaclass=ABCMeta):
    """A Tab which connects to a kernel."""

    kernel: "Kernel"
    kernel_language: "str"
    _metadata: "dict[str, Any]"

    default_callbacks: "MsgCallbacks"
    allow_stdin: "bool"

    def __init__(
        self,
        app: "BaseApp",
        path: "Optional[UPath]" = None,
        kernel: "Optional[Kernel]" = None,
        comms: "Optional[dict[str, Comm]]" = None,
        use_kernel_history: "bool" = False,
    ) -> "None":
        """Create a new instance of a tab with a kernel."""
        super().__init__(app, path)

        if kernel:
            self.kernel = kernel
            self.kernel.default_callbacks = self.default_callbacks
        else:
            self.kernel = Kernel(
                kernel_tab=self,
                allow_stdin=self.allow_stdin,
                default_callbacks=self.default_callbacks,
            )
        self.comms: "dict[str, Comm]" = comms or {}  # The client-side comm states
        self.completer: "Completer" = KernelCompleter(self.kernel)
        self.use_kernel_history = use_kernel_history
        self.history: "History" = (
            KernelHistory(self.kernel) if use_kernel_history else InMemoryHistory()
        )
        self.suggester: "AutoSuggest" = HistoryAutoSuggest(self.history)

    def interrupt_kernel(self) -> "None":
        """Interrupt the current `Notebook`'s kernel."""
        self.kernel.interrupt()

    def restart_kernel(self) -> "None":
        """Restarts the current `Notebook`'s kernel."""
        if confirm := self.app.dialogs.get("confirm"):
            confirm.show(
                message="Are you sure you want to restart the kernel?",
                cb=self.kernel.restart,
            )
        else:
            self.kernel.restart()

    def kernel_started(self, result: "Optional[dict[str, Any]]" = None) -> "None":
        """Tasks to run when the kernel has started."""
        # Check kernel has not failed
        if not self.kernel_name or self.kernel.missing or self.kernel.status == "error":
            return

        # Wait for an idle kernel
        if self.kernel.status != "idle":
            self.kernel.wait_for_status("idle")

        # Load widget comm info
        # self.kernel.comm_info(target_name="jupyter.widget")

        # Load kernel info
        self.kernel.info(set_kernel_info=self.set_kernel_info)

        # Load kernel history
        if self.use_kernel_history:
            self.app.create_background_task(self.load_history())

        # Run queued kernel tasks when the kernel is idle
        log.debug("Running %d kernel tasks", len(self.kernel_queue))
        while self.kernel_queue:
            self.kernel_queue.popleft()()

        self.app.invalidate()

    async def load_history(self) -> "None":
        """Load kernel history."""
        try:
            await self.history.load().__anext__()
        except StopAsyncIteration:
            pass

    @property
    def metadata(self) -> "dict[str, Any]":
        """Return a dictionary to hold notebook / kernel metadata."""
        return self._metadata

    @property
    def kernel_name(self) -> "str":
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get(
            "name", self.app.config.kernel_name
        )

    @kernel_name.setter
    def kernel_name(self, value: "str") -> "None":
        """Return the name of the kernel defined in the notebook JSON."""
        self.metadata.setdefault("kernelspec", {})["name"] = value

    @property
    def language(self) -> "str":
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("language")

    @property
    def kernel_display_name(self) -> "str":
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("display_name", self.kernel_name)

    @property
    def kernel_lang_file_ext(self) -> "str":
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("language_info", {}).get("file_extension", ".py")

    def set_kernel_info(self, info: "dict") -> "None":
        """Handle kernel info requests."""
        self.metadata["language_info"] = info.get("language_info", {})

    def change_kernel(
        self, msg: "Optional[str]" = None, startup: "bool" = False
    ) -> None:
        """Prompt the user to select a new kernel."""
        kernel_specs = self.kernel.specs

        # Warn the user if no kernels are installed
        if not kernel_specs:
            if startup and "no-kernels" in self.app.dialogs:
                self.app.dialogs["no-kernels"].show()
            return

        # Automatically select the only kernel if there is only one
        if startup and len(kernel_specs) == 1:
            self.kernel.change(list(kernel_specs)[0])
            return

        self.app.dialogs["change-kernel"].show(
            tab=self, message=msg, kernel_specs=kernel_specs
        )

    def comm_open(self, content: "dict", buffers: "Sequence[bytes]") -> "None":
        """Register a new kernel Comm object in the notebook."""
        comm_id = str(content.get("comm_id"))
        self.comms[comm_id] = open_comm(
            comm_container=self, content=content, buffers=buffers
        )

    def comm_msg(self, content: "dict", buffers: "Sequence[bytes]") -> "None":
        """Respond to a Comm message from the kernel."""
        comm_id = str(content.get("comm_id"))
        if comm := self.comms.get(comm_id):
            comm.process_data(content.get("data", {}), buffers)

    def comm_close(self, content: "dict", buffers: "Sequence[bytes]") -> "None":
        """Close a notebook Comm."""
        comm_id = content.get("comm_id")
        if comm_id in self.comms:
            del self.comms[comm_id]

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=kernel_tab_has_focus)
    def _change_kernel() -> "None":
        """Change the notebook's kernel."""
        if isinstance(kt := get_app().tab, KernelTab):
            kt.change_kernel()

    # ################################### Settings ####################################

    add_setting(
        name="kernel_name",
        flags=["--kernel-name"],
        type_=str,
        help_="The name of the kernel to start by default",
        default="python3",
        description="""
            The name of the kernel selected automatically by the console app or in new
            notebooks. If set to an empty string, the user will be asked which kernel
            to launch.
        """,
    )



"""Utility functions for formatting key-bindings."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.key_binding.key_bindings import _parse_key

from euporie.core.keys import Keys

if TYPE_CHECKING:
    from typing import Union

    from euporie.core.key_binding.registry import AnyKeys


KEY_ALIASES: "dict[Union[str, Keys], str]" = {
    Keys.ControlH: "backspace",
    Keys.ControlM: "enter",
    Keys.ControlI: "tab",
    Keys.ControlUnderscore: "c-/",
    Keys.ControlAt: "c-space",
}


def parse_keys(keys: "AnyKeys") -> "list[tuple[Union[str, Keys], ...]]":
    """Parse a list of keys."""
    output: "list[tuple[Union[str, Keys], ...]]" = []
    if not isinstance(keys, list):
        keys = [keys]
    for key in keys:
        if isinstance(key, Keys):
            output.append((key,))
        elif isinstance(key, tuple):
            output.append(tuple(_parse_key(k) for k in key))
        else:
            output.append((_parse_key(key),))
    return output


def _format_key_str(key: "str") -> "str":
    if key:
        key = key.replace("c-", "Ctrl+").replace("s-", "Shift+")
        parts = key.split("+")
        if parts[-1].isupper():
            parts.insert(-1, "Shift")
            key = "+".join(parts)
    return key.title()


def format_keys(keys: "list[tuple[Union[str, Keys], ...]]") -> "list[str]":
    """Converts a list of tuples of keys to a string."""
    s: "list[str]" = []

    # Add duplicate key aliases to the list
    keys_: "list[tuple[Union[str, Keys], ...]]" = []
    for key in keys:
        if len(key) == 1 and key[0] in KEY_ALIASES:
            keys_.append((KEY_ALIASES[key[0]],))
        keys_.append(key)

    # Format the key text
    for key in keys_:
        if isinstance(key, tuple):
            s.append(
                ", ".join([_format_key_str(k) for k in key])
                .replace("Escape, ", "Alt+")
                .replace("Alt+Ctrl+m", "Ctrl+Alt+Enter")
            )
        elif isinstance(key, str):
            s.append(_format_key_str(key))
        else:
            s.append(_format_key_str(key.value))

    # Remove duplicate entries but keep the order
    keys_ = list(dict(zip(keys, range(len(keys)))).keys())

    return s



"""Defines an application menu."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition, to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import (
    fragment_list_to_text,
    fragment_list_width,
    to_plain_text,
)
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    Float,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import get_cwidth
from prompt_toolkit.widgets.base import Shadow

from euporie.core.app import get_app
from euporie.core.border import HalfBlockOuterGridStyle

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Optional, Sequence

    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.formatted_text.base import (
        AnyFormattedText,
        OneStyleAndTextTuple,
        StyleAndTextTuples,
    )
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

    from euporie.core.app import BaseApp
    from euporie.core.border import GridStyle
    from euporie.core.commands import Command


log = logging.getLogger(__name__)


class MenuItem:
    """A prompt-toolkit compatible menu item with more advanced capabilities.

    It can use a function to generate formatted text to display, display a checkmark if
    a condition is true, and disable the handler if a condition is met.
    """

    def __init__(
        self,
        formatted_text: "AnyFormattedText" = "",
        description: "str" = "",
        separator: "bool" = False,
        handler: "Optional[Callable[[], None]]" = None,
        children: "Optional[list[MenuItem]]" = None,
        shortcut: "AnyFormattedText" = "",
        hidden: "FilterOrBool" = False,
        disabled: "FilterOrBool" = False,
        toggled: "Optional[Filter]" = None,
        collapse_prefix: "bool" = False,
        collapse_suffix: "bool" = True,
    ) -> "None":
        """Initiate a smart menu item.

        Args:
            formatted_text: A formatted text, or a callable which returns the text to
                display
            description: More information about what the menu item does
            separator: If True, this menu item is treated as a separator
            handler: As per `prompt_toolkit.widgets.menus.MenuItem`
            children: As per `prompt_toolkit.widgets.menus.MenuItem`
            shortcut: As per `prompt_toolkit.widgets.menus.MenuItem`
            hidden: The handler will be hidden when this filter is True
            disabled: The handler will be disabled when this filter is True
            toggled: A checkmark will be displayed next to the menu text when this
                callable returns True
            collapse_prefix: If :py:const:`False`, all prefixes in the menu will be
                padded so they have equal widths
            collapse_suffix: if :py:const:`False`, all suffixes in the menu will be
                padded so they have equal widths

        """
        self._formatted_text = formatted_text
        self.description = description
        self.separator = separator
        self._disabled = to_filter(disabled) | to_filter(self.separator)
        self.hidden = to_filter(hidden)
        self.toggled = toggled
        self.collapse_prefix = collapse_prefix
        self.collapse_suffix = collapse_suffix
        self._prefix_width: "Optional[int]" = None

        self.handler = handler
        self.children = children or []
        self.shortcut = shortcut
        self.selected_item = 0

    @property
    def formatted_text(self) -> "StyleAndTextTuples":
        """Generate the formatted text for this menu item."""
        if callable(self._formatted_text):
            text = self._formatted_text()
        else:
            text = self._formatted_text

        return to_formatted_text(text)

    # Type checking disabled for the following property methods due to open mypy bug:
    # https://github.com/python/mypy/issues/4125

    @property  # type: ignore
    def text(self) -> "str":  # type: ignore
        """Return plain text verision of the item's formatted text."""
        return fragment_list_to_text(self.formatted_text)

    @text.setter
    def text(self, value: "Any") -> "None":
        """Prevent the inherited `__init__` method setting this property value."""
        pass

    @classmethod
    def from_command(cls, command: "Command") -> "MenuItem":
        """Create a menu item from a command."""
        return cls(
            formatted_text=command.menu_title,
            handler=command.menu_handler,
            shortcut=command.key_str,
            disabled=~command.filter,
            hidden=command.hidden,
            toggled=command.toggled,
            description=command.description,
        )

    @property  # type: ignore
    def disabled(self) -> "bool":  # type: ignore
        """Determine if the menu item is disabled."""
        return self._disabled()

    @disabled.setter
    def disabled(self, value: "Any") -> "None":
        """Prevent the inherited `__init__` method setting this property value."""
        pass

    @property
    def has_toggles(self) -> "bool":
        """Returns true if any child items have a toggle state."""
        toggles = False
        for child in self.children:
            if (
                not toggles
                and isinstance(child, MenuItem)
                and child.toggled is not None
            ):
                toggles = True
        return toggles

    @property
    def prefix(self) -> "StyleAndTextTuples":
        """The item's prefix.

        Formatted text that will be displayed before the item's main text. All prefixes
        in a menu are left aligned and padded to take up equal width.

        Returns:
            Formatted text

        """
        prefix: "StyleAndTextTuples" = []
        if self.toggled is not None:
            prefix.append(("class:menu,prefix", "✓ " if self.toggled() else "  "))
        return prefix

    @property
    def prefix_width(self) -> "int":
        """The maximum width of the item's children's prefixes."""
        if self._prefix_width is None:
            self._prefix_width = max(
                [
                    fragment_list_width(child.prefix)
                    if isinstance(child, MenuItem)
                    else 0
                    for child in self.children
                ]
                + [0]
            )

        return self._prefix_width or 0

    @property
    def suffix(self) -> "StyleAndTextTuples":
        """The item's suffix.

        Formatted text that will be displayed aligned right after them item's main text.

        Returns:
            Formatted text

        """
        suffix: "StyleAndTextTuples" = []
        if self.children:
            suffix.append(("", "›"))
        elif self.shortcut is not None:
            suffix += [("class:menu", "  ")]
            suffix += to_formatted_text(self.shortcut, style="class:menu,shortcut")
        return suffix

    @property
    def suffix_width(self) -> "int":
        """The maximum width of the item's children's suffixes."""
        return max(
            [
                fragment_list_width(child.suffix) if isinstance(child, MenuItem) else 0
                for child in self.children
            ]
            + [0]
        )

    @property
    def width(self) -> "int":
        """The maximum width of the item's children."""
        widths = [0]
        for child in self.children:
            width = 0
            if self.collapse_prefix:
                width += (
                    get_cwidth(to_plain_text(child.prefix))
                    if isinstance(child, MenuItem)
                    else 0
                )
            else:
                width += self.prefix_width
            width += get_cwidth(child.text)
            if self.collapse_suffix:
                width += (
                    get_cwidth(to_plain_text(child.suffix))
                    if isinstance(child, MenuItem)
                    else 0
                )
            else:
                width += self.suffix_width
            widths.append(width)
        return max(widths)


class MenuBar:
    """A container to hold the menubar and main application body."""

    def _get_menu(self, level: "int") -> "MenuItem":
        menu = self.menu_items[self.selected_menu[0]]

        for i, index in enumerate(self.selected_menu[1:]):
            if i < level:
                try:
                    menu = menu.children[index]
                except IndexError:
                    return MenuItem("debug")

        return menu

    def _get_menu_fragments(self) -> "StyleAndTextTuples":

        focused = get_app().layout.has_focus(self.window)

        # This is called during the rendering. When we discover that this
        # widget doesn't have the focus anymore. Reset menu state.
        if not focused:
            self.selected_menu = [0]

        def mouse_handler(index: "int", mouse_event: MouseEvent) -> "None":
            hover = mouse_event.event_type == MouseEventType.MOUSE_MOVE
            if mouse_event.event_type == MouseEventType.MOUSE_DOWN or hover and focused:
                # Toggle focus.
                app = get_app()
                if not hover:
                    if app.layout.has_focus(self.window):
                        if self.selected_menu == [index]:
                            app.layout.focus_last()
                    else:
                        app.layout.focus(self.window)
                self.selected_menu = [index]

        results: "StyleAndTextTuples" = []
        used_keys = set()

        for i, item in enumerate(self.menu_items):

            # Add shortcut key hints
            key = to_plain_text(item.formatted_text)[0].lower()
            ft: "StyleAndTextTuples"
            if key not in used_keys:
                ft = explode_text_fragments(item.formatted_text)
                ft = [(f"underline {ft[0][0]}", ft[0][1]), *ft[1:]]
                used_keys |= {key}
            else:
                ft = item.formatted_text

            mh = partial(mouse_handler, i)
            selected = i == self.selected_menu[0] and focused
            style = "class:selection" if selected else ""
            first_style = f"{style} [SetMenuPosition]" if selected else style

            results.extend(
                [
                    (first_style, " ", mh),
                    *[(f"{style} {style_}", text, mh) for style_, text, *_ in ft],
                    (style, " ", mh),
                ]
            )

        return results

    def _submenu(self, level: "int" = 0) -> "Window":
        def get_text_fragments() -> "StyleAndTextTuples":
            result: "StyleAndTextTuples" = []
            if level < len(self.selected_menu):
                menu = self._get_menu(level)

                if menu.children:
                    result.extend(
                        [
                            ("class:menu,border", self.grid.TOP_LEFT),
                            ("class:menu,border", self.grid.TOP_MID * menu.width),
                            ("class:menu,border", self.grid.TOP_RIGHT),
                            ("", "\n"),
                        ]
                    )

                    try:
                        selected_item = self.selected_menu[level + 1]
                    except IndexError:
                        selected_item = -1

                    def one_item(
                        i: int, item: "MenuItem"
                    ) -> "Iterable[OneStyleAndTextTuple]":
                        assert isinstance(item, MenuItem)
                        assert isinstance(menu, MenuItem)

                        def mouse_handler(mouse_event: "MouseEvent") -> "None":
                            if item.disabled:
                                # The arrow keys can't interact with menu items that
                                # are disabled. The mouse shouldn't be able to either.
                                return
                            hover = mouse_event.event_type == MouseEventType.MOUSE_MOVE
                            if (
                                mouse_event.event_type == MouseEventType.MOUSE_UP
                                or hover
                            ):
                                app = get_app()
                                if not hover and item.handler:
                                    app.layout.focus_last()
                                    item.handler()
                                else:
                                    self.selected_menu = self.selected_menu[
                                        : level + 1
                                    ] + [i]

                        if item.separator:
                            # Show a connected line with no mouse handler
                            yield (
                                "class:menu,border",
                                self.grid.SPLIT_LEFT
                                + (self.grid.SPLIT_MID * menu.width)
                                + self.grid.SPLIT_RIGHT,
                            )

                        else:
                            # Show the right edge
                            style = ""
                            # Set the style if disabled
                            if item.disabled:
                                style += "class:menu,disabled"
                            # Set the style and cursor if selected
                            if i == selected_item:
                                style += "class:menu,selection"
                            yield (f"{style} class:menu,border", self.grid.MID_LEFT)
                            if i == selected_item:
                                yield ("[SetCursorPosition]", "")
                            # Construct the menu item contents
                            prefix_padding = " " * (
                                0
                                if menu.collapse_prefix
                                else menu.prefix_width
                                - fragment_list_width(item.prefix)
                            )
                            suffix_padding = " " * (
                                menu.width
                                - fragment_list_width(item.prefix)
                                - len(prefix_padding)
                                - fragment_list_width(item.formatted_text)
                                - (
                                    fragment_list_width(item.suffix)
                                    if menu.collapse_suffix
                                    else menu.suffix_width
                                )
                            )
                            text_padding = " " * (
                                menu.width
                                - fragment_list_width(item.prefix)
                                - len(prefix_padding)
                                - fragment_list_width(item.formatted_text)
                                - fragment_list_width(item.suffix)
                                - len(suffix_padding)
                            )
                            menu_formatted_text: "StyleAndTextTuples" = (
                                to_formatted_text(
                                    [
                                        *item.prefix,
                                        ("", prefix_padding),
                                        *item.formatted_text,
                                        ("", text_padding),
                                        ("", suffix_padding),
                                        *item.suffix,
                                    ],
                                    style=style,
                                )
                            )
                            # Apply mouse handler to all fragments
                            menu_formatted_text = [
                                (fragment[0], fragment[1], mouse_handler)
                                for fragment in menu_formatted_text
                            ]
                            # Show the menu item contents
                            yield from menu_formatted_text
                            # Position the sub-menu
                            if i == selected_item:
                                yield ("[SetMenuPosition]", "")
                            # Show the right edge
                            yield (f"{style} class:menu,border", self.grid.MID_RIGHT)

                        yield ("", "\n")

                    for i, item in enumerate(menu.children):
                        if not item.hidden():
                            result.extend(one_item(i, item))

                    result.extend(
                        [
                            ("class:menu,border", self.grid.BOTTOM_LEFT),
                            ("class:menu,border", self.grid.BOTTOM_MID * menu.width),
                            ("class:menu,border", self.grid.BOTTOM_RIGHT),
                        ]
                    )

            return result

        return Window(FormattedTextControl(get_text_fragments), style="class:menu")

    def __init__(
        self,
        app: "BaseApp",
        menu_items: "Sequence[MenuItem]",
        grid: "GridStyle" = HalfBlockOuterGridStyle,
    ) -> "None":
        """Initiate the menu bar.

        Args:
            app: The application the menubar is attached to
            menu_items: The menu items to show in the menubar
            grid: The grid style to use for the menu's borders

        """
        self.app = app
        self.menu_items = menu_items
        self.grid = grid
        self.selected_menu = [0]

        # Key bindings.
        kb = KeyBindings()

        @Condition
        def in_main_menu() -> bool:
            return len(self.selected_menu) == 1

        @Condition
        def in_sub_menu() -> bool:
            return len(self.selected_menu) > 1

        # Navigation through the main menu.

        @kb.add("left", filter=in_main_menu)
        def _left(event: "KeyPressEvent") -> "None":
            self.selected_menu[0] = max(0, self.selected_menu[0] - 1)

        @kb.add("right", filter=in_main_menu)
        def _right(event: "KeyPressEvent") -> "None":
            self.selected_menu[0] = min(
                len(self.menu_items) - 1, self.selected_menu[0] + 1
            )

        @kb.add("down", filter=in_main_menu)
        def _down(event: "KeyPressEvent") -> "None":
            self.selected_menu.append(0)

        @kb.add("c-c", filter=in_main_menu)
        @kb.add("c-g", filter=in_main_menu)
        def _cancel(event: "KeyPressEvent") -> "None":
            """Leave menu."""
            event.app.layout.focus_last()

        # Sub menu navigation.

        @kb.add("left", filter=in_sub_menu)
        @kb.add("c-g", filter=in_sub_menu)
        @kb.add("c-c", filter=in_sub_menu)
        def _back(event: "KeyPressEvent") -> "None":
            """Go back to parent menu."""
            if len(self.selected_menu) > 1:
                self.selected_menu.pop()

        @kb.add("right", filter=in_sub_menu)
        def _submenu(event: "KeyPressEvent") -> "None":
            """Go into a sub menu."""
            if self._get_menu(len(self.selected_menu) - 1).children:
                self.selected_menu.append(0)

            # If This item does not have a sub menu. Go up in the parent menu.
            elif (
                len(self.selected_menu) == 2
                and self.selected_menu[0] < len(self.menu_items) - 1
            ):
                self.selected_menu = [
                    min(len(self.menu_items) - 1, self.selected_menu[0] + 1)
                ]
                if self.menu_items[self.selected_menu[0]].children:
                    self.selected_menu.append(0)

        @kb.add("up", filter=in_sub_menu)
        def _up_in_submenu(event: "KeyPressEvent") -> "None":
            """Select previous (enabled) menu item or return to main menu."""
            # Look for previous enabled items in this sub menu.
            menu = self._get_menu(len(self.selected_menu) - 2)
            index = self.selected_menu[-1]

            previous_indexes = [
                i
                for i, item in enumerate(menu.children)
                if i < index and not item.disabled
            ]

            if previous_indexes:
                self.selected_menu[-1] = previous_indexes[-1]
            elif len(self.selected_menu) == 2:
                # Return to main menu.
                self.selected_menu.pop()

        @kb.add("down", filter=in_sub_menu)
        def _down_in_submenu(event: "KeyPressEvent") -> "None":
            """Select next (enabled) menu item."""
            menu = self._get_menu(len(self.selected_menu) - 2)
            index = self.selected_menu[-1]

            next_indexes = [
                i
                for i, item in enumerate(menu.children)
                if i > index and not item.disabled
            ]

            if next_indexes:
                self.selected_menu[-1] = next_indexes[0]

        @kb.add("enter")
        def _click(event: "KeyPressEvent") -> "None":
            """Click the selected menu item."""
            item = self._get_menu(len(self.selected_menu) - 1)
            if item.handler:
                event.app.layout.focus_last()
                item.handler()

        @kb.add("escape")
        def _close(event: "KeyPressEvent") -> "None":
            """Close the current menu."""
            if len(self.selected_menu) > 1:
                self.selected_menu = self.selected_menu[:-1]
            else:
                app.layout.focus_previous()

        # Add global CUA menu shortcut
        @kb.add("f10", is_global=True)
        def _open_menu_default(event: "KeyPressEvent") -> "None":
            self.selected_menu = [0]
            event.app.layout.focus(self.window)

        # Add menu shortcuts
        used_keys = set()
        for i, item in enumerate(menu_items):
            key = to_plain_text(item.formatted_text)[0].lower()
            if key not in used_keys:
                used_keys |= {key}

                @kb.add("escape", key, is_global=True)
                def _open_menu(event: "KeyPressEvent", index: "int" = i) -> "None":
                    """Open the  menu item."""
                    self.selected_menu = [index]
                    event.app.layout.focus(self.window)

        # Controls.
        self.control = FormattedTextControl(
            self._get_menu_fragments,
            key_bindings=kb,
            focusable=True,
            show_cursor=False,
        )
        self.window: "Window" = Window(
            height=1,
            content=self.control,
            style="class:menu",
        )

        submenu = self._submenu(0)
        submenu2 = self._submenu(1)
        submenu3 = self._submenu(2)

        @Condition
        def has_focus() -> bool:
            return get_app().layout.current_window == self.window

        self.app.menus["menu-1"] = Float(
            xcursor=True,
            ycursor=True,
            content=ConditionalContainer(
                content=Shadow(body=submenu), filter=has_focus
            ),
        )
        self.app.menus["menu-2"] = Float(
            attach_to_window=submenu,
            xcursor=True,
            ycursor=True,
            allow_cover_cursor=True,
            content=ConditionalContainer(
                content=Shadow(body=submenu2),
                filter=has_focus & Condition(lambda: len(self.selected_menu) >= 1),
            ),
        )
        self.app.menus["menu-3"] = Float(
            attach_to_window=submenu2,
            xcursor=True,
            ycursor=True,
            allow_cover_cursor=True,
            content=ConditionalContainer(
                content=Shadow(body=submenu3),
                filter=has_focus & Condition(lambda: len(self.selected_menu) >= 2),
            ),
        )

        get_app().container_statuses[self.window] = self.statusbar_fields

    def statusbar_fields(
        self,
    ) -> "tuple[list[AnyFormattedText], list[AnyFormattedText]]":
        """Return the description of the currently selected menu item."""
        selected_item = self._get_menu(len(self.selected_menu) - 1)
        if isinstance(selected_item, MenuItem):
            return (["", selected_item.description], [])
        else:
            return (["", ""], [])

    def __pt_container__(self) -> "Container":
        return self.window



"""Defines filters used in the notebook app."""

from __future__ import annotations

from prompt_toolkit.filters import Condition

from euporie.core.app import get_app


@Condition
def notebook_has_focus() -> "bool":
    """Determine if there is a currently focused notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    return isinstance(get_app().tab, Notebook)


@Condition
def deleted_cells() -> "bool":
    """Determine if there ares cell in the undo buffer."""
    from euporie.notebook.tabs.notebook import Notebook

    nb = get_app().tab
    if isinstance(nb, Notebook):
        return bool(nb.undo_buffer)
    return False


@Condition
def code_cell_selected() -> "bool":
    """Determine if a code cell is selected."""
    from euporie.notebook.tabs.notebook import Notebook

    nb = get_app().tab
    if isinstance(nb, Notebook):
        for cell in nb.cells:
            if cell.cell_type == "code":
                return True
    return False


@Condition
def cell_has_focus() -> "bool":
    """Determine if there is a currently focused cell."""
    from euporie.notebook.tabs.notebook import Notebook

    nb = get_app().tab
    if isinstance(nb, Notebook):
        return nb.cell is not None
    return False


@Condition
def in_edit_mode() -> "bool":
    """Determine if there is a currently focused notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    nb = get_app().tab
    if isinstance(nb, Notebook):
        return nb.edit_mode
    return False



"""Contains the main class for a notebook file."""

from __future__ import annotations

import logging
from typing import AsyncGenerator, Iterable

from prompt_toolkit.completion.base import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document

from euporie.core.kernel import Kernel

log = logging.getLogger(__name__)


class KernelCompleter(Completer):
    """A prompt_toolkit completer which provides completions from a Jupyter kernel."""

    def __init__(self, kernel: Kernel) -> None:
        """Instantiate the completer for a given notebook.

        Args:
            kernel: A `Notebook` instance

        """
        self.kernel = kernel

    def get_completions(
        self, document: "Document", complete_event: "CompleteEvent"
    ) -> "Iterable[Completion]":
        """Does nothing as completions are retrieved asynchronously."""
        while False:
            yield

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> "AsyncGenerator[Completion, None]":
        """Retrieves completions from a :class:`Kernel`."""
        for kwargs in await self.kernel.complete_(
            code=document.text,
            cursor_pos=document.cursor_position,
        ):
            if completion_type := kwargs.get("display_meta"):
                kwargs["style"] = f"class:completion-menu.completion.{completion_type}"
                kwargs[
                    "selected_style"
                ] = f"class:completion-menu.completion.current.{completion_type}"
            yield Completion(**kwargs)
